import os
from raichu.visualize_cluster import *
from raichu.modules_to_structure import cluster_to_structure
from raichu.reactions.thioesterase_reactions import thioesterase_all_products
matplotlib.use("Agg")


def export_tabular(genes, file_name="cluster_text.txt"):
    cluster = get_raichu_cluster(genes)
    gene_names = []
    pks = False
    nrps = False
    for gene in genes:
        for module in gene.modules:
            gene_names.append(gene.name)
            if module.type == 'NRPS':
                nrps = True
            if module.type == 'PKS':
                pks = True

    if nrps and pks:
        cluster_type = 'NRPS/PKS hybrid'
    elif pks:
        cluster_type = 'PKS type I'
    elif nrps:
        cluster_type = 'NRPS'
    else:
        cluster_type = ''

    with open(file_name, 'w') as out:
        out.write('bgc_type\tgene_name\tmodule_id\tmodule_type\tspecificity\ttailoring_domains\n')
        for i, module in enumerate(cluster):
            if len(module) == 4:
                out.write(f"{cluster_type}\t{gene_names[i]}\t{i}\t{module[1]}\t{module[2]}\t{'|'.join(module[3])}\n")
            else:
                out.write(f"{cluster_type}\t{gene_names[i]}\t{i}\t{module[1]}\t{module[2]}\n")


def get_raichu_cluster(genes):
    cluster = []

    module_nr = 0

    if genes:

        for i, gene in enumerate(genes):
            last_gene = False

            if gene.name:
                file_name = f"{gene}.png"
            else:
                file_name = f"gene_{i}.png"

            for module in gene.modules:
                substrate_specificity = None
                module_nr += 1

                module_name = f"module {module_nr}"

                if module.type == 'NRPS' and module_nr == 1:
                    module_type = 'starter_module_nrps'

                elif module.type == 'NRPS':
                    module_type = 'elongation_module_nrps'
                elif module.type == 'PKS' and module_nr == 1:
                    module_type = 'starter_module_pks'
                    substrate_specificity = 'SC(=O)CC'
                    for domain in module.domains:
                        if domain.type == 'AT':
                            if domain.substrate:
                                substrate_specificity = domain.substrate.smiles

                else:
                    module_type = 'elongation_module_pks'

                    substrate_specificity = 'pk'
                    for domain in module.domains:
                        if domain.type == 'AT':
                            if domain.substrate:
                                substrate_specificity = domain.substrate.name

                    if substrate_specificity == 'methoxymalonylcoa':
                        substrate_specificity = 'methoxymalonylacp'
                    if substrate_specificity == 'wildcard':
                        substrate_specificity = 'pk'

                if module.type == 'NRPS':
                    substrate_specificity = 'nrp'
                    for domain in module.domains:
                        if domain.type == 'A':
                            if domain.substrate:
                                if domain.substrate.custom:
                                    substrate_specificity = domain.substrate.smiles
                                else:
                                    substrate_specificity = domain.substrate.name

                assert substrate_specificity

                last_domain = False

                tailoring_domains = []
                for domain in module.domains:
                    if domain.type not in ['A', 'C', 'AT', 'KR', 'KS', 'ACP', 'PCP', 'TE']:
                        tailoring_domains.append(domain.type)
                    elif domain.type == 'KR':
                        if domain.subtype:
                            tailoring_domains = [f'KR_{domain.subtype}'] + tailoring_domains
                        else:
                            tailoring_domains = ["KR"] + tailoring_domains
                    elif domain.type == 'TE':
                        last_domain = True
                        if module.type == 'NRPS':
                            module_type = 'terminator_module_nrps'
                        elif module.type == 'PKS':

                            module_type = 'terminator_module_pks'

                if not gene.name:
                    gene_name = f"gene {i + 1}"
                else:
                    gene_name = gene.name

                if module_type == "starter_module_pks":

                    cluster.append([module_name, module_type, substrate_specificity, gene_name])
                else:
                    cluster.append([module_name, module_type, substrate_specificity, tailoring_domains, gene_name])

                if last_domain:
                    last_gene = True
                    break

            if last_gene:
                break

    return cluster


def render_products(genes, dir_name="products_test"):

    cluster = get_raichu_cluster(genes)

    if cluster:
        if not os.path.exists(dir_name):
            os.mkdir(dir_name)
        attached_product = cluster_to_structure(cluster, attach_to_acp=True)
        thioesterase_all_products(attached_product, out_folder=dir_name)


def render_cluster(genes, file_name="cluster_test.png"):
    cluster = get_raichu_cluster(genes)
    if cluster:
        draw_cluster(cluster, save_fig=file_name)




import os

from pikachu.chem.structure import Structure
from pikachu.reactions.functional_groups import find_bonds, find_atoms
from pikachu.reactions.basic_reactions import hydrolysis, internal_condensation
from pikachu.general import structure_to_smiles

from raichu.data.molecular_moieties import SC_BOND, O_OH, O_BETAPROPRIOLACTONE_O,\
    O_BETAPROPRIOLACTONE_TERMINAL_O
from raichu.reactions.general import initialise_atom_attributes
from raichu.drawing.drawer import RaichuDrawer


def release_linear_reduction(chain_intermediate: Structure) -> Structure:
    sc_bonds = find_bonds(SC_BOND, chain_intermediate)
    if len(sc_bonds) != 1:
        raise ValueError("Cannot release product from carrier domain as no SC bond is present")

    sc_bond = sc_bonds[0]

    carbon = sc_bond.get_neighbour('C')
    sulphur = sc_bond.get_neighbour('S')

    chain_intermediate.break_bond(sc_bond)

    chain_intermediate.add_atom('H', [carbon])
    chain_intermediate.add_atom('H', [sulphur])

    initialise_atom_attributes(chain_intermediate)

    structures = chain_intermediate.split_disconnected_structures()
    for structure in structures:
        if carbon in structure.graph:
            structure.refresh_structure()
            return structure

    raise RuntimeError("Could not release chain through reduction.")


def release_linear_thioesterase(chain_intermediate):
    """Carries out the thioesterase reaction on the polyketide chain
    intermediate, returning the free acid linear polyketide product as a
    PIKAChU Structure object

    chain_intermediate: PIKAChU Structure object of the polyketide chain
    intermediate, either attached to a PKS domain or not
    """
    # Find S-H bond in chain intermediate and break the bond and define atoms
    sc_bonds = find_bonds(SC_BOND, chain_intermediate)
    assert len(sc_bonds) == 1
    carbon = sc_bonds[0].get_neighbour('C')

    linear_product = None

    structures = hydrolysis(chain_intermediate, sc_bonds[0])
    for structure in structures:
        if carbon in structure.graph:
            linear_product = structure
            break

    assert linear_product

    initialise_atom_attributes(linear_product)

    carbon = linear_product.get_atom(carbon)
    carbon.annotations.set_annotation('terminal_c', True)

    oxygens = carbon.get_neighbours('O')
    terminal_oxygen = None

    for oxygen in oxygens:
        if oxygen.has_neighbour('H'):
            terminal_oxygen = oxygen
            break

    assert terminal_oxygen

    terminal_oxygen.annotations.set_annotation('terminal_o', True)

    return linear_product


def cyclic_release(linear_product, o_oh_n_amino):
    """Performs the thioesterase reactions on the input chain_intermediate
     using the -OH group defined by the input O-atom participating in that
     internal -OH group, returns the circular product as PIKAChU Structure
     object.

     chain_intermediate: PIKAChU Structure object of a polyketide
     o_oh_n_amino: PIKAChU Atom object of the O-atom in the -OH group or N-atom
     in the amino group that the function should use to perform the
     thioesterase reaction.
    """

    # Remove H from internal -OH group and refresh structure
    cyclisation_site = linear_product.get_atom(o_oh_n_amino)
    h_atom = cyclisation_site.get_neighbour('H')
    assert h_atom
    h_bond = cyclisation_site.get_bond(h_atom)
    assert h_bond
    terminal_oxygen = None
    terminal_carbon = None

    for atom in linear_product.graph:
        if atom.annotations.terminal_o:
            terminal_oxygen = atom
        elif atom.annotations.terminal_c:
            terminal_carbon = atom

    assert terminal_oxygen and terminal_carbon

    oh_bond = terminal_oxygen.get_bond(terminal_carbon)

    cyclic_product, water = internal_condensation(linear_product, oh_bond, h_bond)

    return cyclic_product


def find_o_betapropriolactone(polyketide):
    """
    Finds and returns the oxygen atom (PIKAChU atom object) in the -OH group
    that shouldn't be used by the thioesterase_circular_product function, as
    this will create a beta-propriolactone compound, which does not occur in
    polyketide synthesis, if present. Otherwise the function returns None

    polyketide: PIKAChU structure object of a polyketide
    """
    o_propriolactone = find_atoms(O_BETAPROPRIOLACTONE_O, polyketide)
    o_propriolactone_terminal = find_atoms(O_BETAPROPRIOLACTONE_TERMINAL_O, polyketide)

    assert len(o_propriolactone) == len(o_propriolactone_terminal)

    o_beta_propriolactones = []

    for i, atom in enumerate(o_propriolactone):
        terminal_o = o_propriolactone_terminal[i]
        if terminal_o.annotations.terminal_o:
            o_beta_propriolactones.append(atom)

    if len(o_beta_propriolactones) == 0:
        return None
    elif len(o_beta_propriolactones) == 1:
        return o_beta_propriolactones[0]
    else:
        raise ValueError('Error: this molecule is not a polyketide, as the \
        carbon in the beta ketone/hydroxyl group is bound to an additional \
         oxygen atom')


def thioesterase_all_products(chain_intermediate, out_folder=None):
    """Performs all thioesterase reactions on the input chain_intermediate
     using all internal amino and -OH groups except for the -OH group that
     leads to the formation of a beta-propriolactone compound, which does not
     occur in polyketide synthesis. Returns a list of PIKAChU Structure objects
     of all possible thioesterase products.

     chain_intermediate: PIKAChU Structure object of a polyketide/NRP
    """
    # Perform first thioesterase reaction, generating linear polyketide/NRP

    chain_intermediate_copy = chain_intermediate.deepcopy()
    linear_product = release_linear_thioesterase(chain_intermediate_copy)
    if not out_folder:
        RaichuDrawer(linear_product)

    # Find OH groups in polyketide/NRP, perform cyclization for each -OH group
    o_oh_atoms = find_atoms(O_OH, linear_product)

    o_oh_atoms_filtered = []
    for atom in o_oh_atoms:
        if not atom.annotations.terminal_o and atom not in o_oh_atoms_filtered and atom.has_neighbour('H'):
            o_oh_atoms_filtered.append(atom)

    # Find amino groups in polyketide/NRP, perform cyclization for each group

    amino_n_atoms_filtered = []

    for atom in linear_product.graph:
        if atom.type == 'N':
            if len(atom.get_neighbours('H')) == 2 and atom.has_neighbour('C') and not atom.aromatic and atom \
                    not in amino_n_atoms_filtered:
                amino_n_atoms_filtered.append(atom)

    # Define -OH group that should not be used to carry out the thioesterase
    # reaction (distance -S and internal -OH group)
    o_not_to_use = find_o_betapropriolactone(linear_product)

    list_product_drawings = []
    circular_smiles = []

    # Perform all possible thioesterase reactions leading to the formation of
    # circular products using the internal amino groups, save Structure objects
    # to list

    for n_amino in amino_n_atoms_filtered:
        linear_product_copy = linear_product.deepcopy()
        n_atom = linear_product_copy.get_atom(n_amino)
        product = cyclic_release(linear_product_copy, n_atom)
        circular_smiles.append(structure_to_smiles(product))
        list_product_drawings.append(product)

    # Perform all possible thioesterase reactions leading to the formation of
    # circular products using the internal -OH groups, save Structure objects
    # to list
    for o_oh in o_oh_atoms_filtered:
        linear_product_copy = linear_product.deepcopy()
        oh_atom = linear_product_copy.get_atom(o_oh)

        if oh_atom != o_not_to_use:
            product = cyclic_release(linear_product_copy, oh_atom)
            circular_smiles.append(structure_to_smiles(product))
            list_product_drawings.append(product)

    if out_folder:
        smiles_path = os.path.join(out_folder, "product_smiles.txt")
        smiles_file = open(smiles_path, 'w')

        file_path = os.path.join(out_folder, f"product_0.png")

        if os.path.exists(file_path):
            os.remove(file_path)

        linear_smiles = structure_to_smiles(linear_product)
        smiles_file.write(f"product_0\t{linear_smiles}\n")

        # png_from_smiles(linear_smiles, file_path)

        drawing = RaichuDrawer(linear_product, save_png=file_path)
        drawing.draw_structure()

        for i, product in enumerate(list_product_drawings):
            smiles = circular_smiles[i]
            smiles_file.write(f"product_{i}\t{smiles}\n")

            file_path = os.path.join(out_folder, f"product_{i + 1}.png")
            if os.path.exists(file_path):
                os.remove(file_path)
            drawing = RaichuDrawer(product, save_png=file_path)
            drawing.draw_structure()

            # png_from_smiles(smiles, file_path)

        smiles_file.close()

    else:

        # Draw all products
        for product in list_product_drawings:
            RaichuDrawer(product)

    return list_product_drawings



from pikachu.general import read_smiles, draw_structure
from pikachu.reactions.functional_groups import combine_structures

from raichu.data.attributes import ATTRIBUTES


def epimerization(chiral_centre):
    new_chirality = None
    if chiral_centre.chiral == 'clockwise':
        new_chirality = 'counterclockwise'
    elif chiral_centre.chiral == 'counterclockwise':
        new_chirality = 'clockwise'

    chiral_centre.chiral = new_chirality


def epimerize(nrp):
    """

    """
    did_reaction = False
    # Define epimerization reaction target
    chiral_c_ep_atoms = []
    for atom in nrp.graph:
        if atom.annotations.chiral_c_ep:
            chiral_c_ep_atoms.append(atom)

    # If the substrate is a (fatty) acid, epimerization is not possible
    if len(chiral_c_ep_atoms) == 0:
       print('Warning: Cannot perform epimerization reaction on non-amino acid substrate!')

    assert len(chiral_c_ep_atoms) < 2

    # Carry out epimerization
    if len(chiral_c_ep_atoms) == 1:
        chiral_c = chiral_c_ep_atoms[0]
        epimerization(chiral_c)
        if chiral_c.chiral:
            did_reaction = True

    return nrp, did_reaction


def methylation(target_atom, structure):

    methyl_group = read_smiles('C')
    methyl_group.add_attributes(ATTRIBUTES, boolean=True)
    carbon = methyl_group.atoms[0]
    hydrogen_1 = methyl_group.atoms[1]
    bond_1 = carbon.get_bond(hydrogen_1)

    hydrogen_2 = target_atom.get_neighbour('H')
    if not hydrogen_2:
        raise Exception("Can't methylate this atom!")

    bond_2 = target_atom.get_bond(hydrogen_2)

    methylated_structure = combine_structures([structure, methyl_group])

    methylated_structure.break_bond(bond_1)
    methylated_structure.break_bond(bond_2)

    methylated_structure.make_bond(carbon, target_atom, methylated_structure.find_next_bond_nr())
    methylated_structure.make_bond(hydrogen_1, hydrogen_2, methylated_structure.find_next_bond_nr())

    structures = methylated_structure.split_disconnected_structures()

    for s in structures:
        if carbon.nr in s.atoms:
            return s


def n_methylate(nrp):
    """

    """
    # Define N-methylation reaction target
    n_meth_locations = []
    for atom in nrp.graph:
        if atom.annotations.n_atom_nmeth:
            n_meth_locations.append(atom)

    # If the substrate is not an amino acid, N-methylation is not possible
    if len(n_meth_locations) == 0:
        print('Warning: Cannot perform N-methylation on a non-amino acid substrate!')

    assert len(n_meth_locations) < 2

    product = None

    if len(n_meth_locations) == 1:

        n_meth = n_meth_locations[0]

        # Check if the N atom has a hydrogen group necessary for the reaction, and
        # not a cyclic amiono acid such as proline
        executable = True
        if not n_meth.has_neighbour('H'):
            executable = False
            product = nrp

        # Carry out N-methylation
        if executable:
            product = methylation(n_meth, nrp)

    else:
        product = nrp

    assert product

    return product, True


if __name__ == "__main__":
    s = read_smiles(r"NCC(=O)NCC(=O)O")
    target_atom = s.atoms[4]
    structure = methylation(target_atom, s)
    draw_structure(structure)








from raichu.data.attributes import ATTRIBUTES


def label_rest_groups(subunit, chain_intermediate):
    atom_types = []

    for atom in chain_intermediate.graph:
        atom_types.append(atom.type)

    nr_unknown_atoms = atom_types.count('*')

    counter = 0
    for atom in subunit.graph:
        if atom.type == '*':
            counter += 1
            # TODO: replace with atom attribute
            atom.unknown_index = nr_unknown_atoms + counter


def initialise_atom_attributes(structure):
    for atom in structure.graph:
        if not hasattr(atom.annotations, 'in_central_chain'):

            for attribute in ATTRIBUTES:
                atom.annotations.add_annotation(attribute, False)


def reset_nrp_annotations(structure):
    for atom in structure.graph:
        atom.annotations.chiral_c_ep = False
        atom.annotations.n_atom_nmeth = False
        atom.annotations.c2_acid = False
        atom.annotations.leaving_oh_h = False
        atom.annotations.leaving_oh_o = False



from pikachu.drawing.drawing import *
from pikachu.math_functions import *
from pikachu.smiles.smiles import Smiles

from raichu.central_chain_detection.find_central_chain import find_central_chain


class RaichuDrawer(Drawer):
    def __init__(self, structure, options=None, save_png=None, dont_show=False,
                 coords_only=True, dpi=100, save_svg=None):
        self.dont_show = dont_show
        self.dpi = dpi
        if options is None:
            self.options = Options()
        else:
            self.options = options
        if save_png is None:
            self.save_png = None
        else:
            # Check if filename is valid
            assert save_png.endswith('.png')
            self.save_png = save_png
        if save_svg is None:
            self.save_svg = None
        else:
            # Check if filename is valid
            assert save_svg.endswith('.svg')
            self.save_svg = save_svg
        super().__init__(structure, options=None, coords_only=coords_only)

    def center_on_carrier_domain(self):
        pass

    def export_to_json(self):
        json_structure = {"atoms": [],
                          "bonds": [],
                          "modules": []}

        ring_centers_x = []
        ring_centers_y = []

        for ring in self.rings:
            self.set_ring_center(ring)

            ring_centers_x.append(ring.center.x)
            ring_centers_y.append(ring.center.y)

        for bond_nr, bond in self.structure.bonds.items():
            if bond.atom_1.draw.positioned and bond.atom_2.draw.positioned:
                json_bond = {"id": bond.nr,
                             "atom_1": bond.atom_1.nr,
                             "atom_2": bond.atom_2.nr,
                             "type": bond.type,
                             "chiral": False,
                             "lines": []}

                if bond in self.chiral_bonds:

                    orientation, chiral_center = self.chiral_bond_to_orientation[bond]
                    json_bond["chiral"] = True
                    json_bond["wedge_orientation"] = orientation
                    json_bond["wedge_origin"] = chiral_center.nr
                elif (bond.atom_1.type == 'S' and bond.atom_2.annotations.domain_type or \
                      (bond.atom_2.type == 'S' and bond.atom_1.annotations.domain_type)):
                    json_bond["type"] = "squiggle"

        # If the starter unit contains an unknown moiety, number this R1
        for atom in self.structure.graph:
            if atom.type == '*' and not atom.annotations.unknown_index:
                atom.annotations.unknown_index = 1
        for atom in self.structure.graph:
            if atom.draw.positioned:
                json_atom = {"id": atom.nr,
                             "type": atom.type,
                             "text": '',
                             "atom_x": atom.draw.position.x,
                             "atom_y": atom.draw.position.y}

                if atom.type != 'C' and atom.draw.positioned:
                    text_h = ''
                    text_h_pos = None

                    if atom.type != 'C' or atom.draw.draw_explicit:
                        text = atom.type
                    else:
                        text = ''

                    if atom.annotations.domain_type:
                        text = atom.annotations.domain_type

                    horizontal_alignment = 'center'

                    orientation = self.get_hydrogen_text_orientation(atom)
                    if orientation == 'H_above_atom':
                        text_h_pos = Vector(atom.draw.position.x, atom.draw.position.y + 6)
                    if orientation == 'H_below_atom':
                        text_h_pos = Vector(atom.draw.position.x, atom.draw.position.y - 6)

                    atom_draw_position = Vector(atom.draw.position.x, atom.draw.position.y)

                    if atom.type == '*':
                        neighbouring_c = atom.get_neighbour('C')
                        assert neighbouring_c
                        # In order to not let the number of the sidechain overlap
                        # with the bond, move Rx symbol along hor. axis depending on
                        # if group is added to the left or right of the chain
                        delta_x_r = self.get_delta_x_sidechain(atom, neighbouring_c)
                        atom_draw_position.x += delta_x_r
                        text = fr'$R_{atom.annotations.unknown_index}$'

                    if not atom.charge and (atom.type != 'C' or atom.draw.draw_explicit):

                        if atom.draw.has_hydrogen:
                            hydrogen_count = 0
                            for neighbour in atom.neighbours:
                                if neighbour.type == 'H' and not neighbour.draw.is_drawn:
                                    hydrogen_count += 1

                            if hydrogen_count:

                                if hydrogen_count > 1:
                                    if orientation == 'H_before_atom':
                                        text = r'$H_{hydrogens}{atom_type}$'.format(hydrogens=hydrogen_count,
                                                                                    atom_type=atom.type)
                                        horizontal_alignment = 'right'
                                        atom_draw_position.x += 3
                                    elif orientation == 'H_below_atom' or orientation == 'H_above_atom':
                                        text = atom.type
                                        text_h = r'$H_{hydrogens}$'.format(hydrogens=hydrogen_count)

                                    else:
                                        text = r'${atom_type}H_{hydrogens}$'.format(hydrogens=hydrogen_count,
                                                                                    atom_type=atom.type)
                                        horizontal_alignment = 'left'
                                        atom_draw_position.x -= 3
                                elif hydrogen_count == 1:
                                    if orientation == 'H_before_atom':
                                        text = f'H{atom.type}'
                                        horizontal_alignment = 'right'
                                        atom_draw_position.x += 3
                                    elif orientation == 'H_below_atom' or orientation == 'H_above_atom':
                                        text = atom.type
                                        text_h = 'H'
                                    else:
                                        text = f'{atom.type}H'
                                        horizontal_alignment = 'left'
                                        atom_draw_position.x -= 3

                    elif atom.charge:
                        if atom.charge > 0:
                            charge_symbol = '+'
                        else:
                            charge_symbol = '-'

                        hydrogen_count = 0
                        for neighbour in atom.neighbours:
                            if neighbour.type == 'H' and not neighbour.draw.is_drawn:
                                hydrogen_count += 1

                        if not hydrogen_count:

                            if abs(atom.charge) > 1:

                                text = r'${atom_type}^{charge}{charge_symbol}$'.format(charge=atom.charge,
                                                                                       atom_type=atom.type,
                                                                                       charge_symbol=charge_symbol)
                            elif abs(atom.charge) == 1:
                                text = r'${atom_type}^{charge_symbol}$'.format(atom_type=atom.type,
                                                                               charge_symbol=charge_symbol)

                            horizontal_alignment = 'left'
                            atom_draw_position.x -= 3
                        else:

                            if hydrogen_count > 1:
                                if orientation == 'H_before_atom':
                                    if abs(atom.charge) > 1:
                                        text = r'$H_{hydrogens}{atom_type}^{charge}{charge_symbol}$'.format(
                                            hydrogens=hydrogen_count,
                                            atom_type=atom.type,
                                            charge=atom.charge,
                                            charge_symbol=charge_symbol)
                                    elif abs(atom.charge) == 1:
                                        text = r'$H_{hydrogens}{atom_type}^{charge_symbol}$'.format(
                                            hydrogens=hydrogen_count,
                                            atom_type=atom.type,
                                            charge_symbol=charge_symbol)

                                    horizontal_alignment = 'right'
                                    atom_draw_position.x += 3
                                elif orientation == 'H_above_atom' or orientation == 'H_below_atom':
                                    text_h = r'$H_{hydrogens}$'.format(hydrogens=hydrogen_count)
                                    if abs(atom.charge) > 1:
                                        text = r'${atom_type}^{charge}{charge_symbol}$'.format(atom_type=atom.type,
                                                                                               charge=atom.charge,
                                                                                               charge_symbol=charge_symbol)
                                    elif abs(atom.charge) == 1:
                                        text = r'${atom_type}^{charge_symbol}$'.format(atom_type=atom.type,
                                                                                       charge_symbol=charge_symbol)
                                else:
                                    if abs(atom.charge) > 1:
                                        text = r'${atom_type}H_{hydrogens}^{charge}{charge_symbol}$'.format(
                                            hydrogens=hydrogen_count,
                                            atom_type=atom.type,
                                            charge=atom.charge,
                                            charge_symbol=charge_symbol)
                                    elif abs(atom.charge) == 1:
                                        text = r'${atom_type}H_{hydrogens}^{charge_symbol}$'.format(
                                            hydrogens=hydrogen_count,
                                            atom_type=atom.type,
                                            charge_symbol=charge_symbol)

                                    horizontal_alignment = 'left'
                                    atom_draw_position.x -= 3
                            elif hydrogen_count == 1:
                                if orientation == 'H_before_atom':
                                    if abs(atom.charge) > 1:

                                        text = r'$H{atom_type}^{charge}{charge_symbol}$'.format(atom_type=atom.type,
                                                                                                charge=atom.charge,
                                                                                                charge_symbol=charge_symbol)
                                    elif abs(atom.charge) == 1:
                                        text = r'$H{atom_type}^{charge_symbol}$'.format(atom_type=atom.type,
                                                                                        charge_symbol=charge_symbol)
                                    horizontal_alignment = 'right'
                                    atom_draw_position.x += 3
                                elif orientation == 'H_above_atom' or orientation == 'H_below_atom':
                                    text_h = 'H'
                                    if abs(atom.charge) > 1:

                                        text = r'${atom_type}^{charge}{charge_symbol}$'.format(atom_type=atom.type,
                                                                                               charge=atom.charge,
                                                                                               charge_symbol=charge_symbol)
                                    elif abs(atom.charge) == 1:
                                        text = r'${atom_type}^{charge_symbol}$'.format(atom_type=atom.type,
                                                                                       charge_symbol=charge_symbol)

                                else:
                                    if abs(atom.charge) > 1:
                                        text = r'${atom_type}H^{charge}{charge_symbol}$'.format(atom_type=atom.type,
                                                                                                charge=atom.charge,
                                                                                                charge_symbol=charge_symbol)

                                    elif abs(atom.charge) == 1:
                                        text = r'${atom_type}H^{charge_symbol}$'.format(atom_type=atom.type,
                                                                                        charge_symbol=charge_symbol)
                                    horizontal_alignment = 'left'
                                    atom_draw_position.x -= 3

                    if text:
                        plt.text(atom_draw_position.x, atom_draw_position.y,
                                 text,
                                 horizontalalignment=horizontal_alignment,
                                 verticalalignment='center',
                                 color=atom.draw.colour)
                    if text_h:
                        plt.text(text_h_pos.x, text_h_pos.y,
                                 text_h,
                                 horizontalalignment='center',
                                 verticalalignment='center',
                                 color=atom.draw.colour)

        # If a png filename is included in the initialization of the
        # Raichu_drawer object, don't show the structure, but do save it as a
        # png image to the provided filename
        if self.dont_show:
            return self

        elif self.save_png is None and not self.dont_show and self.save_svg is None:
            plt.show()

        else:
            if self.save_png:
                plt.savefig(self.save_png)
            elif self.save_svg:
                plt.savefig(self.save_svg)
            plt.clf()
            plt.close()

    def finetune_overlap_resolution(self, masked_bonds=None, highest_atom=None):

        if not masked_bonds:
            masked_bonds = set()

        else:
            masked_bonds = set(masked_bonds)

        if self.total_overlap_score > self.options.overlap_sensitivity:
            clashing_atoms = self.find_clashing_atoms()

            best_bonds = []
            for atom_1, atom_2 in clashing_atoms:
                shortest_path = self.find_shortest_path(atom_1, atom_2)
                rotatable_bonds = []
                distances = []

                for i, bond in enumerate(shortest_path):
                    distance_1 = i
                    distance_2 = len(shortest_path) - i

                    average_distance = len(shortest_path) / 2

                    distance_metric = abs(average_distance - distance_1) + abs(average_distance - distance_2)

                    if self.bond_is_rotatable(bond) and bond not in masked_bonds:
                        rotatable_bonds.append(bond)
                        distances.append(distance_metric)

                best_bond = None
                optimal_distance = float('inf')
                for i, distance in enumerate(distances):
                    if distance < optimal_distance:
                        if highest_atom:
                            if not atom_1.draw.position.x > highest_atom.draw.position.x and atom_2.draw.position.x > highest_atom.draw.position.x:
                                best_bond = rotatable_bonds[i]
                                optimal_distance = distance
                        else:
                            best_bond = rotatable_bonds[i]
                            optimal_distance = distance

                if best_bond:
                    best_bonds.append(best_bond)

            best_bonds = list(set(best_bonds))

            for best_bond in best_bonds:
                if self.total_overlap_score > self.options.overlap_sensitivity:

                    subtree_size_1 = self.get_subgraph_size(best_bond.atom_1, {best_bond.atom_2})
                    subtree_size_2 = self.get_subgraph_size(best_bond.atom_2, {best_bond.atom_1})

                    if subtree_size_1 < subtree_size_2:
                        rotating_atom = best_bond.atom_1
                        parent_atom = best_bond.atom_2
                    else:
                        rotating_atom = best_bond.atom_2
                        parent_atom = best_bond.atom_1

                    overlap_score, _, _ = self.get_overlap_score()

                    scores = [overlap_score]

                    for i in range(12):
                        self.rotate_subtree(rotating_atom, parent_atom, math.radians(30), parent_atom.draw.position)
                        new_overlap_score, _, _ = self.get_overlap_score()
                        scores.append(new_overlap_score)

                    assert len(scores) == 13

                    scores = scores[:12]

                    best_i = 0
                    best_score = scores[0]

                    for i, score in enumerate(scores):
                        if score < best_score:
                            best_score = score
                            best_i = i

                    self.total_overlap_score = best_score

                    self.rotate_subtree(rotating_atom, parent_atom, math.radians(30 * best_i + 1), parent_atom.draw.position)

    def draw_structure(self):
        min_x = 100000000
        max_x = -100000000
        min_y = 100000000
        max_y = -100000000

        for atom in self.structure.graph:
            if atom.draw.positioned:
                if atom.draw.position.x < min_x:
                    min_x = atom.draw.position.x
                if atom.draw.position.y < min_y:
                    min_y = atom.draw.position.y
                if atom.draw.position.x > max_x:
                    max_x = atom.draw.position.x
                if atom.draw.position.y > max_y:
                    max_y = atom.draw.position.y

        height = max_y - min_y
        width = max_x - min_x
        self.line_width = 2

        fig, ax = plt.subplots(figsize=((width + 2 * self.options.padding) /
        50.0, (height + 2 * self.options.padding) / 50.0), dpi=self.dpi)

        ax.set_aspect('equal', adjustable='box')
        ax.axis('off')

        ax.set_xlim(
            [min_x - self.options.padding, max_x + self.options.padding])
        ax.set_ylim(
            [min_y - self.options.padding, max_y + self.options.padding])
        plt.subplots_adjust(left=0, bottom=0, right=1, top=1, wspace=0,
                            hspace=0)

        params = {'mathtext.default': 'regular', }
        plt.rcParams.update(params)

        ring_centers_x = []
        ring_centers_y = []

        for ring in self.rings:
            self.set_ring_center(ring)

            ring_centers_x.append(ring.center.x)
            ring_centers_y.append(ring.center.y)

        for bond_nr, bond in self.structure.bonds.items():
            if bond.atom_1.draw.positioned and bond.atom_2.draw.positioned:
                line = Line(bond.atom_1.draw.position,
                            bond.atom_2.draw.position, bond.atom_1,
                            bond.atom_2)
                midpoint = line.get_midpoint()
                truncated_line = line.get_truncated_line(
                    self.options.short_bond_length)
                if bond.type == 'single':
                    if bond in self.chiral_bonds:
                        orientation, chiral_center = self.chiral_bond_to_orientation[bond]
                        self.plot_chiral_bond(orientation, chiral_center, line,
                                              ax, midpoint)
                    else:
                        if (bond.atom_1.type == 'S' and
                                bond.atom_2.annotations.domain_type or
                                (bond.atom_2.type == 'S' and
                                 bond.atom_1.annotations.domain_type)):
                            self.plot_halflines_s_domain(line, ax, midpoint)
                        else:
                            self.plot_halflines(line, ax, midpoint)
                elif bond.type == 'double':
                    if not self.is_terminal(
                            bond.atom_1) and not self.is_terminal(bond.atom_2):
                        self.plot_halflines(line, ax, midpoint)

                        common_ring_numbers = self.get_common_rings(
                            bond.atom_1, bond.atom_2)

                        if common_ring_numbers:
                            common_rings = []
                            for ring_nr in common_ring_numbers:
                                common_rings.append(self.get_ring(ring_nr))

                            common_rings.sort(key=lambda x: len(x.members))
                            common_ring = common_rings[0]
                            ring_centre = common_ring.center
                            second_line = line.double_line_towards_center(
                                ring_centre, self.options.bond_spacing,
                                self.options.double_bond_length)
                            second_line_midpoint = second_line.get_midpoint()
                            self.plot_halflines_double(second_line, ax,
                                                       second_line_midpoint)

                        else:
                            bond_neighbours = bond.atom_1.drawn_neighbours +\
                                              bond.atom_2.drawn_neighbours
                            if bond_neighbours:
                                vectors = [atom.draw.position for atom in
                                           bond_neighbours]
                                gravitational_point = Vector.get_average(
                                    vectors)
                                second_line = line.double_line_towards_center(
                                    gravitational_point,
                                    self.options.bond_spacing,
                                    self.options.double_bond_length)
                                second_line_midpoint = second_line.get_midpoint()
                                self.plot_halflines_double(second_line, ax,
                                                           second_line_midpoint)
                            else:
                                print("Shouldn't happen!")
                    else:
                        if self.is_terminal(bond.atom_1) and self.is_terminal(
                                bond.atom_2):
                            dummy_1 = Vector(bond.atom_1.draw.position.x + 1,
                                             bond.atom_1.draw.position.y + 1)
                            dummy_2 = Vector(bond.atom_1.draw.position.x - 1,
                                             bond.atom_1.draw.position.y - 1)
                            double_bond_line_1 = line.double_line_towards_center(
                                dummy_1,
                                self.options.bond_spacing / 2.0,
                                self.options.double_bond_length)
                            double_bond_line_1_midpoint = double_bond_line_1.get_midpoint()
                            double_bond_line_2 = line.double_line_towards_center(
                                dummy_2,
                                self.options.bond_spacing / 2.0,
                                self.options.double_bond_length)
                            double_bond_line_2_midpoint = double_bond_line_2.get_midpoint()

                            self.plot_halflines_double(double_bond_line_1, ax,
                                                       double_bond_line_1_midpoint)
                            self.plot_halflines_double(double_bond_line_2, ax,
                                                       double_bond_line_2_midpoint)

                        else:

                            if self.is_terminal(bond.atom_1):
                                terminal_atom = bond.atom_1
                                branched_atom = bond.atom_2
                            else:
                                terminal_atom = bond.atom_2
                                branched_atom = bond.atom_1

                            if len(branched_atom.drawn_neighbours) >= 3:
                                closest_two = self.get_sorted_distances_from_list(
                                    terminal_atom,
                                    branched_atom.drawn_neighbours)
                                closest_atom_1 = closest_two[0][1]
                                closest_atom_2 = closest_two[1][1]

                                line = Line(terminal_atom.draw.position,
                                            branched_atom.draw.position,
                                            terminal_atom, branched_atom)

                                bond_1_line = Line(branched_atom.draw.position,
                                                   closest_atom_1.draw.position,
                                                   branched_atom,
                                                   closest_atom_1)
                                bond_2_line = Line(branched_atom.draw.position,
                                                   closest_atom_2.draw.position,
                                                   branched_atom,
                                                   closest_atom_1)

                                double_bond_line_1 = line.double_line_towards_center(
                                    closest_atom_1.draw.position,
                                    self.options.bond_spacing / 2.0,
                                    self.options.double_bond_length)
                                double_bond_line_2 = line.double_line_towards_center(
                                    closest_atom_2.draw.position,
                                    self.options.bond_spacing / 2.0,
                                    self.options.double_bond_length)

                                double_bond_line_1_midpoint = \
                                    double_bond_line_1.get_midpoint()
                                double_bond_line_2_midpoint = \
                                    double_bond_line_2.get_midpoint()

                                intersection_1 = double_bond_line_1.find_intersection(
                                    bond_1_line)
                                intersection_2 = double_bond_line_2.find_intersection(
                                    bond_2_line)

                                if terminal_atom.draw.position.x >\
                                        branched_atom.draw.position.x:
                                    double_bond_line_1.point_1 = intersection_1
                                    double_bond_line_2.point_1 = intersection_2
                                else:
                                    double_bond_line_1.point_2 = intersection_1
                                    double_bond_line_2.point_2 = intersection_2

                                self.plot_halflines(double_bond_line_1, ax,
                                                    double_bond_line_1_midpoint)
                                self.plot_halflines(double_bond_line_2, ax,
                                                    double_bond_line_2_midpoint)

                            else:
                                pass

        # If the starter unit contains an unknown moiety, number this R1
        for atom in self.structure.graph:
            if atom.type == '*' and not atom.annotations.unknown_index:
                atom.annotations.unknown_index = 1
        for atom in self.structure.graph:
            if atom.type != 'C' and atom.draw.positioned:
                text_h = ''
                text_h_pos = None

                if atom.type != 'C' or atom.draw.draw_explicit:
                    text = atom.type
                else:
                    text = ''

                if atom.annotations.domain_type:
                    text = atom.annotations.domain_type

                horizontal_alignment = 'center'

                orientation = self.get_hydrogen_text_orientation(atom)
                if orientation == 'H_above_atom':
                    text_h_pos = Vector(atom.draw.position.x, atom.draw.position.y + 6)
                if orientation == 'H_below_atom':
                    text_h_pos = Vector(atom.draw.position.x, atom.draw.position.y - 6)

                atom_draw_position = Vector(atom.draw.position.x, atom.draw.position.y)

                if atom.type == '*':
                    neighbouring_c = atom.get_neighbour('C')
                    assert neighbouring_c
                    # In order to not let the number of the sidechain overlap
                    # with the bond, move Rx symbol along hor. axis depending on
                    # if group is added to the left or right of the chain
                    delta_x_r = self.get_delta_x_sidechain(atom, neighbouring_c)
                    atom_draw_position.x += delta_x_r
                    text = fr'$R_{atom.annotations.unknown_index}$'

                if not atom.charge and (atom.type != 'C' or atom.draw.draw_explicit):

                    if atom.draw.has_hydrogen:
                        hydrogen_count = 0
                        for neighbour in atom.neighbours:
                            if neighbour.type == 'H' and not neighbour.draw.is_drawn:
                                hydrogen_count += 1

                        if hydrogen_count:

                            if hydrogen_count > 1:
                                if orientation == 'H_before_atom':
                                    text = r'$H_{hydrogens}{atom_type}$'.format(hydrogens=hydrogen_count,
                                                                                atom_type=atom.type)
                                    horizontal_alignment = 'right'
                                    atom_draw_position.x += 3
                                elif orientation == 'H_below_atom' or orientation == 'H_above_atom':
                                    text = atom.type
                                    text_h = r'$H_{hydrogens}$'.format(hydrogens=hydrogen_count)

                                else:
                                    text = r'${atom_type}H_{hydrogens}$'.format(hydrogens=hydrogen_count,
                                                                                atom_type=atom.type)
                                    horizontal_alignment = 'left'
                                    atom_draw_position.x -= 3
                            elif hydrogen_count == 1:
                                if orientation == 'H_before_atom':
                                    text = f'H{atom.type}'
                                    horizontal_alignment = 'right'
                                    atom_draw_position.x += 3
                                elif orientation == 'H_below_atom' or orientation == 'H_above_atom':
                                    text = atom.type
                                    text_h = 'H'
                                else:
                                    text = f'{atom.type}H'
                                    horizontal_alignment = 'left'
                                    atom_draw_position.x -= 3

                elif atom.charge:
                    if atom.charge > 0:
                        charge_symbol = '+'
                    else:
                        charge_symbol = '-'

                    hydrogen_count = 0
                    for neighbour in atom.neighbours:
                        if neighbour.type == 'H' and not neighbour.draw.is_drawn:
                            hydrogen_count += 1

                    if not hydrogen_count:

                        if abs(atom.charge) > 1:

                            text = r'${atom_type}^{charge}{charge_symbol}$'.format(charge=atom.charge,
                                                                                   atom_type=atom.type,
                                                                                   charge_symbol=charge_symbol)
                        elif abs(atom.charge) == 1:
                            text = r'${atom_type}^{charge_symbol}$'.format(atom_type=atom.type,
                                                                           charge_symbol=charge_symbol)

                        horizontal_alignment = 'left'
                        atom_draw_position.x -= 3
                    else:

                        if hydrogen_count > 1:
                            if orientation == 'H_before_atom':
                                if abs(atom.charge) > 1:
                                    text = r'$H_{hydrogens}{atom_type}^{charge}{charge_symbol}$'.format(
                                        hydrogens=hydrogen_count,
                                        atom_type=atom.type,
                                        charge=atom.charge,
                                        charge_symbol=charge_symbol)
                                elif abs(atom.charge) == 1:
                                    text = r'$H_{hydrogens}{atom_type}^{charge_symbol}$'.format(
                                        hydrogens=hydrogen_count,
                                        atom_type=atom.type,
                                        charge_symbol=charge_symbol)

                                horizontal_alignment = 'right'
                                atom_draw_position.x += 3
                            elif orientation == 'H_above_atom' or orientation == 'H_below_atom':
                                text_h = r'$H_{hydrogens}$'.format(hydrogens=hydrogen_count)
                                if abs(atom.charge) > 1:
                                    text = r'${atom_type}^{charge}{charge_symbol}$'.format(atom_type=atom.type,
                                                                                           charge=atom.charge,
                                                                                           charge_symbol=charge_symbol)
                                elif abs(atom.charge) == 1:
                                    text = r'${atom_type}^{charge_symbol}$'.format(atom_type=atom.type,
                                                                                   charge_symbol=charge_symbol)
                            else:
                                if abs(atom.charge) > 1:
                                    text = r'${atom_type}H_{hydrogens}^{charge}{charge_symbol}$'.format(
                                        hydrogens=hydrogen_count,
                                        atom_type=atom.type,
                                        charge=atom.charge,
                                        charge_symbol=charge_symbol)
                                elif abs(atom.charge) == 1:
                                    text = r'${atom_type}H_{hydrogens}^{charge_symbol}$'.format(
                                        hydrogens=hydrogen_count,
                                        atom_type=atom.type,
                                        charge_symbol=charge_symbol)

                                horizontal_alignment = 'left'
                                atom_draw_position.x -= 3
                        elif hydrogen_count == 1:
                            if orientation == 'H_before_atom':
                                if abs(atom.charge) > 1:

                                    text = r'$H{atom_type}^{charge}{charge_symbol}$'.format(atom_type=atom.type,
                                                                                            charge=atom.charge,
                                                                                            charge_symbol=charge_symbol)
                                elif abs(atom.charge) == 1:
                                    text = r'$H{atom_type}^{charge_symbol}$'.format(atom_type=atom.type,
                                                                                    charge_symbol=charge_symbol)
                                horizontal_alignment = 'right'
                                atom_draw_position.x += 3
                            elif orientation == 'H_above_atom' or orientation == 'H_below_atom':
                                text_h = 'H'
                                if abs(atom.charge) > 1:

                                    text = r'${atom_type}^{charge}{charge_symbol}$'.format(atom_type=atom.type,
                                                                                           charge=atom.charge,
                                                                                           charge_symbol=charge_symbol)
                                elif abs(atom.charge) == 1:
                                    text = r'${atom_type}^{charge_symbol}$'.format(atom_type=atom.type,
                                                                                   charge_symbol=charge_symbol)

                            else:
                                if abs(atom.charge) > 1:
                                    text = r'${atom_type}H^{charge}{charge_symbol}$'.format(atom_type=atom.type,
                                                                                            charge=atom.charge,
                                                                                            charge_symbol=charge_symbol)

                                elif abs(atom.charge) == 1:
                                    text = r'${atom_type}H^{charge_symbol}$'.format(atom_type=atom.type,
                                                                                    charge_symbol=charge_symbol)
                                horizontal_alignment = 'left'
                                atom_draw_position.x -= 3

                if text:
                    plt.text(atom_draw_position.x, atom_draw_position.y,
                             text,
                             horizontalalignment=horizontal_alignment,
                             verticalalignment='center',
                             color=atom.draw.colour)
                if text_h:
                    plt.text(text_h_pos.x, text_h_pos.y,
                             text_h,
                             horizontalalignment='center',
                             verticalalignment='center',
                             color=atom.draw.colour)

        # If a png filename is included in the initialization of the
        # Raichu_drawer object, don't show the structure, but do save it as a
        # png image to the provided filename
        if self.dont_show:
            return self

        elif self.save_png is None and not self.dont_show and self.save_svg is None:
            plt.show()

        else:
            if self.save_png:
                plt.savefig(self.save_png)
            elif self.save_svg:
                plt.savefig(self.save_svg)
            plt.clf()
            plt.close()


    def process_structure(self):
        self.position()
        self.structure.refresh_structure()
        self.restore_ring_information()
        self.resolve_primary_overlaps()
        self.total_overlap_score, sorted_overlap_scores, atom_to_scores = self.get_overlap_score()

        for i in range(self.options.overlap_resolution_iterations):
            for bond in self.drawn_bonds:
                if self.bond_is_rotatable(bond):

                    tree_depth_1 = self.get_subgraph_size(bond.atom_1,
                                                          {bond.atom_2})
                    tree_depth_2 = self.get_subgraph_size(bond.atom_2,
                                                          {bond.atom_1})

                    atom_1 = bond.atom_2
                    atom_2 = bond.atom_1

                    if tree_depth_1 > tree_depth_2:
                        atom_1 = bond.atom_1
                        atom_2 = bond.atom_2

                    subtree_overlap_score, _ = self.get_subtree_overlap_score(
                        atom_2, atom_1, atom_to_scores)

                    if subtree_overlap_score > self.options.overlap_sensitivity:
                        neighbours_2 = atom_2.drawn_neighbours[:]
                        neighbours_2.remove(atom_1)

                        if len(neighbours_2) == 1:
                            neighbour = neighbours_2[0]
                            angle = neighbour.draw.position.get_rotation_away_from_vector(
                                atom_1.draw.position, atom_2.draw.position,
                                math.radians(120))

                            self.rotate_subtree(neighbour, atom_2, angle,
                                                atom_2.draw.position)

                            new_overlap_score, _, _ = self.get_overlap_score()
                            if new_overlap_score > self.total_overlap_score:
                                self.rotate_subtree(neighbour, atom_2,
                                                    -angle,
                                                    atom_2.draw.position)
                            else:
                                self.total_overlap_score = new_overlap_score

                        elif len(neighbours_2) == 2:
                            if atom_2.draw.rings and atom_1.draw.rings:
                                continue

                            neighbour_1 = neighbours_2[0]
                            neighbour_2 = neighbours_2[1]
                            if len(neighbour_1.draw.rings) == 1 and len(
                                    neighbour_2.draw.rings) == 1:
                                # If the neighbours are in different rings,
                                # or in rings at all, do nothing
                                if neighbour_1.draw.rings[0] != \
                                        neighbour_2.draw.rings[0]:
                                    continue
                                elif neighbour_1.draw.rings or neighbour_2.draw.rings:
                                    continue
                                else:
                                    angle_1 = neighbour_1.draw.position.get_rotation_away_from_vector(
                                        atom_1.position, atom_2.position,
                                        math.radians(120))
                                    angle_2 = neighbour_2.draw.position.get_rotation_away_from_vector(
                                        atom_1.position, atom_2.position,
                                        math.radians(120))

                                    self.rotate_subtree(neighbour_1,
                                                        atom_2,
                                                        angle_1,
                                                        atom_2.position)
                                    self.rotate_subtree(neighbour_2,
                                                        atom_2,
                                                        angle_2,
                                                        atom_2.position)

                                    new_overlap_score, _, _ = self.get_overlap_score()

                                    if new_overlap_score > self.total_overlap_score:
                                        self.rotate_subtree(neighbour_1,
                                                            atom_2,
                                                            -angle_1,
                                                            atom_2.position)
                                        self.rotate_subtree(neighbour_2,
                                                            atom_2,
                                                            -angle_2,
                                                            atom_2.position)
                                    else:
                                        self.total_overlap_score = new_overlap_score
                        self.total_overlap_score, sorted_overlap_scores, atom_to_scores = self.get_overlap_score()


        # substructure search to see if structure is a polyketide or NRP

        is_polyketide = False
        is_nrp = False
        if self.structure.find_substructures(
                Smiles('SC(=O)').smiles_to_structure()):
            if self.structure.find_substructures(
                    Smiles('SC(CN)=O').smiles_to_structure()):
                is_nrp = True
            else:
                is_polyketide = True

        attached_to_domain = False
        for atom in self.structure.graph:
            if atom.type == 'S':
                sulphur = atom
                for neighbour in sulphur.neighbours:
                    if neighbour.annotations.domain_type:
                        attached_to_domain = True
                        domain = neighbour

        self.structure.refresh_structure()

        ### NRPS + PK code: Force pk/peptide backbone to be drawn straight:
        # If struct=PK/NRP, find central chain and attached domain
        if attached_to_domain and (is_nrp or is_polyketide):
            backbone_atoms = find_central_chain(self.structure)
            pcp = None
            for atom in self.structure.graph:
                if atom.annotations.domain_type:
                    pcp = atom

            assert pcp

            # Fix position of the S and C atom and the S-C angle
            sulphur = backbone_atoms[0]
            first_carbon = backbone_atoms[1]

            sulphur.draw.position.x = first_carbon.draw.position.x
            sulphur.draw.position.y = first_carbon.draw.position.y + 15

            angle = get_angle(first_carbon.draw.position,
                              sulphur.draw.position)
            angle_degrees = round(math.degrees(angle), 3)

            correct_angle_deg = -120
            delta_angle_deg = correct_angle_deg - angle_degrees
            delta_angle_rad = math.radians(delta_angle_deg)

            self.rotate_subtree(sulphur, first_carbon, delta_angle_rad,
                                first_carbon.draw.position)

            # Fix position domain straight above sulphur atom
            pcp.draw.position.x = sulphur.draw.position.x
            pcp.draw.position.y = sulphur.draw.position.y + 15

            # Rotate all other bonds in peptide backbone of NRP
            i = 0
            fixed_atoms = set()
            while i < (len(backbone_atoms) - 1):
                atom1 = backbone_atoms[i]
                atom2 = backbone_atoms[i + 1]
                angle = get_angle(atom1.draw.position, atom2.draw.position)
                angle_degrees = round(math.degrees(angle), 3)

                # Save angle last two atoms, needed later
                if i == (len(backbone_atoms) - 2):
                    last_angle_degrees = angle_degrees

                # Special case for amino acids with cyclic backbone (proline)
                if atom1.inside_ring and atom2.inside_ring:
                    if angle_degrees != 90.0:
                        correct_angle_deg = 90
                        delta_angle_deg = correct_angle_deg - angle_degrees
                        delta_angle_rad = math.radians(delta_angle_deg)
                        self.rotate_subtree(atom2, atom1, delta_angle_rad,
                                            atom1.draw.position)
                    i += 1

                    # Flip cyclic backbone amino acid if necessary
                    if atom1.draw.position.x > backbone_atoms[i - 2].draw.position.x:
                        first_atom_cycle = None
                        for next_atom in atom1.neighbours:
                            print(next_atom, next_atom.type)
                            if next_atom.type != 'H' and next_atom not in backbone_atoms:
                                first_atom_cycle = next_atom
                        assert first_atom_cycle
                        if atom1.draw.position.x > first_atom_cycle.draw.position.x:
                            masked = {atom1, atom2}
                            for atom in self.traverse_substructure(
                                    first_atom_cycle, masked):
                                delta_x = 2 * (atom1.draw.position.x - atom.draw.position.x)
                                atom.draw.position.x += delta_x

                    if atom1.draw.position.x < backbone_atoms[i - 2].draw.position.x:
                        first_atom_cycle = None
                        for next_atom in atom1.neighbours:
                            if next_atom.type != 'H' and next_atom not in backbone_atoms:
                                first_atom_cycle = next_atom
                        assert first_atom_cycle
                        if atom1.draw.position.x < first_atom_cycle.draw.position.x:
                            masked = {atom1, atom2}
                            for atom in self.traverse_substructure(first_atom_cycle, masked):
                                delta_x = 2 * (atom.draw.position.x - atom1.draw.position.x)
                                atom.draw.position.x -= delta_x

                # Fix bond angle backbone atoms if second backbone atom (one
                # with larger position.x) is inside ring, and the first is not
                elif atom2 != backbone_atoms[-1] and atom2.inside_ring and not atom1.inside_ring and \
                        backbone_atoms[i + 2].inside_ring:
                    if atom2 not in fixed_atoms:
                        if round(math.degrees(get_angle(
                                backbone_atoms[i - 1].draw.position,
                                backbone_atoms[i].draw.position)),
                                 3) == 120.0:
                            correct_angle_deg = 60.0
                            first_angle_cyclic = 60.0
                        elif round(math.degrees(get_angle(
                                backbone_atoms[i - 1].draw.position,
                                backbone_atoms[i].draw.position)),
                                   3) == 60.0:
                            correct_angle_deg = 120.0
                            first_angle_cyclic = 120.0
                        else:
                            print('should not happen!!!')
                        delta_angle_deg = correct_angle_deg - angle_degrees
                        delta_angle_rad = math.radians(delta_angle_deg)
                        self.rotate_subtree(atom2, atom1, delta_angle_rad,
                                            atom1.draw.position)

                        fixed_atoms.add(atom2)
                        i = 0
                    else:
                        if angle_degrees == 120.0:
                            first_angle_cyclic = 120.0
                        elif angle_degrees == 60.0:
                            first_angle_cyclic = 60.0
                        i += 1

                # Other way around... (first one in ring, after cycle)
                elif atom1.inside_ring and not atom2.inside_ring and \
                        backbone_atoms[i - 1].inside_ring:
                    if first_angle_cyclic == 60.0:
                        correct_angle_deg = 120.0
                    elif first_angle_cyclic == 120.0:
                        correct_angle_deg = 60.0
                    if angle_degrees != correct_angle_deg:
                        delta_angle_deg = correct_angle_deg - angle_degrees
                        delta_angle_rad = math.radians(delta_angle_deg)
                        self.rotate_subtree(atom2, atom1, delta_angle_rad,
                                            atom1.draw.position)
                        i = 0
                    else:
                        i += 1
                else:
                    if angle_degrees != 120.0 and angle_degrees != 60:
                        if round(math.degrees(get_angle(
                                backbone_atoms[i - 1].draw.position,
                                backbone_atoms[i].draw.position)),
                                 3) == 120.0:
                            correct_angle_deg = 60.0
                        elif round(math.degrees(get_angle(
                                backbone_atoms[i - 1].draw.position,
                                backbone_atoms[i].draw.position)),
                                   3) == 60.0:
                            correct_angle_deg = 120.0
                        delta_angle_deg = correct_angle_deg - angle_degrees
                        delta_angle_rad = math.radians(delta_angle_deg)
                        self.rotate_subtree(atom2, atom1, delta_angle_rad,
                                            atom1.draw.position)
                        i = 0
                    else:
                        if i >= 1:
                            if angle_degrees == 120.0:
                                if round(math.degrees(get_angle(
                                        backbone_atoms[
                                            i - 1].draw.position,
                                        backbone_atoms[i].draw.position)),
                                         3) == 120:
                                    correct_angle_deg = 60.0
                                    delta_angle_deg = correct_angle_deg - angle_degrees
                                    delta_angle_rad = math.radians(
                                        delta_angle_deg)
                                    self.rotate_subtree(atom2, atom1,
                                                        delta_angle_rad,
                                                        atom1.draw.position)
                                    i = 0
                                else:
                                    i += 1
                            elif angle_degrees == 60.0:
                                if round(math.degrees(get_angle(
                                        backbone_atoms[
                                            i - 1].draw.position,
                                        backbone_atoms[i].draw.position)),
                                         3) == 60:
                                    correct_angle_deg = 120.0
                                    delta_angle_deg = correct_angle_deg - angle_degrees
                                    delta_angle_rad = math.radians(
                                        delta_angle_deg)
                                    self.rotate_subtree(atom2, atom1,
                                                        delta_angle_rad,
                                                        atom1.draw.position)
                                    i = 0
                                else:
                                    i += 1
                        else:
                            i += 1

            # Force pk/amino acid sidechains to stick out straight from each side
            i = 1
            while i < (len(backbone_atoms)):
                terminal_carboxylic_acid = False
                atom = backbone_atoms[i]
                atom_neighbours = []
                atom_neighbour_types = []
                connected_to_sidechain = False
                for neighbour in atom.neighbours:
                    atom_neighbours.append(neighbour)
                    atom_neighbour_types.append(neighbour.type)

                    if neighbour not in backbone_atoms and \
                            neighbour.type != 'H' and neighbour.type != 'S' and\
                            neighbour.type == 'O' and \
                            self.structure.bond_lookup[neighbour][atom].type == 'double':
                        first_atom_sidechain = neighbour
                        connected_to_sidechain = True
                        if backbone_atoms[i - 1].draw.position.x < \
                                backbone_atoms[i].draw.position.x:
                            sidechain_orientation = 'right'
                        elif backbone_atoms[i - 1].draw.position.x > \
                                backbone_atoms[i].draw.position.x:
                            sidechain_orientation = 'left'
                    elif neighbour not in backbone_atoms and \
                            neighbour.type != 'H' and neighbour.type != 'S' and\
                            not neighbour.inside_ring:
                        first_atom_sidechain = neighbour
                        connected_to_sidechain = True
                        if backbone_atoms[i - 1].draw.position.x < \
                                backbone_atoms[i].draw.position.x:
                            sidechain_orientation = 'right'
                        elif backbone_atoms[i - 1].draw.position.x > \
                                backbone_atoms[i].draw.position.x:
                            sidechain_orientation = 'left'

                    elif neighbour not in backbone_atoms and \
                            neighbour.type != 'H' and neighbour.type != 'S' and \
                            neighbour.inside_ring and any(
                            bond.type == 'double' for bond in
                            neighbour.bonds) and atom == backbone_atoms[-1]:
                        first_atom_sidechain = neighbour
                        connected_to_sidechain = True
                        if backbone_atoms[i - 1].draw.position.x < \
                                backbone_atoms[i].draw.position.x:
                            sidechain_orientation = 'diagonal_right'
                        elif backbone_atoms[i - 1].draw.position.x > \
                                backbone_atoms[i].draw.position.x:
                            sidechain_orientation = 'diagonal_left'

                    elif neighbour not in backbone_atoms and \
                            neighbour.type != 'H' and neighbour.type != 'S' and \
                            neighbour.in_ring(self.structure) and not atom.in_ring(self.structure):
                        first_atom_sidechain = neighbour
                        connected_to_sidechain = True
                        if backbone_atoms[i - 1].draw.position.x < \
                                backbone_atoms[i].draw.position.x:
                            sidechain_orientation = 'right'
                        elif backbone_atoms[i - 1].draw.position.x > \
                                backbone_atoms[i].draw.position.x:
                            sidechain_orientation = 'left'

                    # If bond is directly connected to sidechain, check&rotate
                    if connected_to_sidechain:
                        angle = get_angle(atom.draw.position,
                                          first_atom_sidechain.draw.position)
                        angle_degrees = round(math.degrees(angle), 3)
                        # Check if it is a terminal carboxylic acid group
                        if atom_neighbour_types.count('O') == 2 and len(
                                atom_neighbour_types) == 3:
                            for next_atom in atom_neighbours:
                                if next_atom.type == 'O':
                                    if \
                                    self.structure.bond_lookup[next_atom][atom].type == 'single':
                                        hydroxyl = next_atom
                                        angle2 = get_angle(
                                            atom.draw.position,
                                            hydroxyl.draw.position)
                                        angle_degrees2 = round(
                                            math.degrees(angle2),
                                            3)
                                        if last_angle_degrees == 120.0:
                                            correct_angle_deg = 60.0
                                        elif last_angle_degrees == 60.0:
                                            correct_angle_deg = 120.0
                                        delta_angle_deg = correct_angle_deg - angle_degrees2
                                        delta_angle_rad = math.radians(
                                            delta_angle_deg)
                                        self.rotate_subtree(hydroxyl,
                                                            atom,
                                                            delta_angle_rad,
                                                            atom.draw.position)
                                    elif \
                                    self.structure.bond_lookup[next_atom][
                                        atom].type == 'double':
                                        carbonyl = next_atom
                                        angle2 = get_angle(
                                            atom.draw.position,
                                            carbonyl.draw.position)
                                        angle_degrees2 = round(
                                            math.degrees(angle2),
                                            3)
                                        correct_angle_deg = 0
                                        delta_angle_deg = correct_angle_deg - angle_degrees2
                                        delta_angle_rad = math.radians(
                                            delta_angle_deg)
                                        self.rotate_subtree(carbonyl, atom,
                                                            delta_angle_rad,
                                                            atom.draw.position)
                                    terminal_carboxylic_acid = True
                        if not terminal_carboxylic_acid:
                            if sidechain_orientation == 'right':
                                correct_angle_deg = 180
                            elif sidechain_orientation == 'left':
                                correct_angle_deg = 0
                            elif sidechain_orientation == 'diagonal_left':
                                correct_angle_deg = 120
                            elif sidechain_orientation == 'diagonal_right':
                                correct_angle_deg = 60
                            delta_angle_deg = correct_angle_deg - angle_degrees
                            delta_angle_rad = math.radians(delta_angle_deg)
                            self.rotate_subtree(first_atom_sidechain, atom,
                                                delta_angle_rad,
                                                atom.draw.position)
                i += 1

            # If the drawer rotated the entire structure, correct this
            angle = get_angle(pcp.draw.position,
                              sulphur.draw.position)
            angle_degrees = round(math.degrees(angle), 3)
            if angle_degrees != 90.0:
                correct_angle_deg = 90
                delta_angle_deg = correct_angle_deg - angle_degrees
                delta_angle_rad = math.radians(delta_angle_deg)
                self.rotate_subtree(sulphur, pcp, delta_angle_rad,
                                    pcp.draw.position)

            # Fix rotation bulky sidechains so carboxyl groups dont need to move
            i = 1
            while i < (len(backbone_atoms)):
                atom = backbone_atoms[i]
                atom_neighbours = []
                atom_neighbour_types = []
                for neighbour in atom.neighbours:
                    atom_neighbours.append(neighbour)
                    atom_neighbour_types.append(neighbour.type)
                    if neighbour not in backbone_atoms and \
                            neighbour.type != 'H' and neighbour.type != 'S' and\
                            not neighbour.inside_ring:
                        first_atom_sidechain = neighbour
                        for further_atom in first_atom_sidechain.neighbours:
                            types = []
                            for further_atom_neighbour in further_atom.neighbours:
                                types.append(further_atom_neighbour.type)
                            if further_atom.type == 'C' and further_atom not in backbone_atoms and (
                                    (types.count(
                                            'C') == 3) or (
                                            types.count('O') == 2 and
                                            types.count('H') == 0) or
                                    (types.count('N') == 1 and
                                     types.count('O') == 1 and
                                     types.count('H') == 0) or
                                    (types.count('C') == 2 and
                                     types.count('N') == 1 and
                                     types.count('H') == 0)):
                                angle_bulky_sidechain = get_angle(
                                    first_atom_sidechain.draw.position,
                                    further_atom.draw.position)
                                angle_bulky_sidechain = round(
                                    math.degrees(angle_bulky_sidechain), 3)
                                if atom.draw.position.x < first_atom_sidechain.draw.position.x:
                                    correct_angle_deg = 160
                                elif further_atom.draw.position.x < first_atom_sidechain.draw.position.x:
                                    correct_angle_deg = 20
                                delta_angle_deg = correct_angle_deg - angle_bulky_sidechain
                                delta_angle_rad = math.radians(
                                    delta_angle_deg)
                                self.rotate_subtree(further_atom,
                                                    first_atom_sidechain,
                                                    delta_angle_rad,
                                                    first_atom_sidechain.draw.position)
                    # Fix incorrect rotation carbonyl groups
                    if neighbour.type == 'O' and \
                            self.structure.bond_lookup[atom][
                                neighbour].type == 'double':
                        correctly_rotated = True
                        if neighbour.draw.position.x > atom.draw.position.x and atom.draw.position.x < \
                                backbone_atoms[i - 1].draw.position.x:
                            correct_angle_deg = 0
                            correctly_rotated = False
                        elif neighbour.draw.position.y != atom.draw.position.y and atom.draw.position.x < \
                                backbone_atoms[i - 1].draw.position.x:
                            correct_angle_deg = 0
                            correctly_rotated = False
                        elif neighbour.draw.position.x < atom.draw.position.x and atom.draw.position.x > \
                                backbone_atoms[i - 1].draw.position.x:
                            correct_angle_deg = 180
                            correctly_rotated = False
                        elif neighbour.draw.position.y != atom.draw.position.y and atom.draw.position.x > \
                                backbone_atoms[i - 1].draw.position.x:
                            correct_angle_deg = 180
                            correctly_rotated = False
                        if not correctly_rotated:
                            angle = get_angle(atom.draw.position,
                                              neighbour.draw.position)
                            angle_degrees = round(math.degrees(angle), 3)
                            delta_angle_deg = correct_angle_deg - angle_degrees
                            delta_angle_rad = math.radians(delta_angle_deg)
                            self.rotate_subtree(neighbour,
                                                atom, delta_angle_rad,
                                                atom.draw.position)
                i += 1

            # self.resolve_primary_overlaps()
            self.total_overlap_score, sorted_overlap_scores, atom_to_scores = self.get_overlap_score()
            central_chain_bonds = set()
            for bond in self.structure.bonds.values():
                if bond.atom_1.annotations.in_central_chain or\
                        bond.atom_2.annotations.in_central_chain:
                    central_chain_bonds.add(bond)

            self.finetune_overlap_resolution(masked_bonds=central_chain_bonds, highest_atom=sulphur)
            self.resolve_secondary_overlaps(sorted_overlap_scores)

    # End NRPS rotation code

    def plot_halflines_s_domain(self, line, ax, midpoint):
        truncated_line = line.get_truncated_line(self.options.short_bond_length)
        self.plot_line_dashed(truncated_line, ax, color='#a6a6a6')

    def plot_line_dashed(self, line, ax, color='grey'):
        with matplotlib.rc_context({'path.sketch': (5, 10, 1)}):
            ax.plot([line.point_1.x, line.point_2.x],
                    [line.point_1.y, line.point_2.y], color=color,
                    linewidth=self.line_width/1.5)

    def get_delta_x_sidechain(self, sidechain_atom, neighbouring_carbon):
        """Returns the change tot the x-position of the sidegroup symbol in
        the structure drawing as float
        sidechain_atom: Atom object of the unknown sidechain group
        neighoburin_carbon: Atom object of the carbon in the central chain
        bound to the unknown sidechain group
        """
        sidechainatom_x = sidechain_atom.draw.position.x
        carbon_x = neighbouring_carbon.draw.position.x
        if (carbon_x - sidechainatom_x) < -10:
            delta_x = 1.5
        elif (carbon_x - sidechainatom_x) > 10:
            delta_x = -1.5
        else:
            delta_x = 0.0

        return delta_x

def get_angle(vector1, vector2):
    difference = Vector.subtract_vectors(vector1, vector2)
    difference_angle = difference.angle()
    return difference_angle



ATTRIBUTES = ['in_central_chain', 'KR_ep_target', 'KR_red_target',
              'latest_elongation_o', 'latest_elongation_methyl', 'DH_target',
              'ER_target', 'domain_type', 'unknown_index', 'chiral_c_ep',
              'n_atom_nmeth', 'leaving_oh_o', 'leaving_oh_h', 'c2_acid',
              'terminal_c', 'terminal_o']

ALL_PKS_ELONGATION_UNITS = ['malonylcoa', 'methylmalonylcoa',
                            'methoxymalonylacp', 'ethylmalonylcoa', 'pk']




from typing import Tuple

from pikachu.reactions.functional_groups import BondDefiner, GroupDefiner, find_atoms, find_bonds, combine_structures
from pikachu.reactions.basic_reactions import internal_condensation
from pikachu.chem.chirality import same_chirality
from pikachu.chem.structure import Structure
from raichu.domain.domain_types import KRDomainSubtype, ERDomainSubtype
from raichu.reactions.general import initialise_atom_attributes
from pikachu.general import read_smiles
from raichu.data.attributes import ATTRIBUTES

FIRST_C = GroupDefiner('first_c', r'CC(S)=O', 1)
RECENT_ALPHA_C = GroupDefiner('recent_alpha_c', r'CCC(S)=O', 1)
RECENT_BETA_C = GroupDefiner('recent_beta_c', r'CCC(S)=O', 0)
RECENT_BETA_C_OH = GroupDefiner('recent_beta_c_oh', r'OCCC(S)=O', 0)
RECENT_ALPHA_METHYL_C = GroupDefiner('recent_alpha_methyl_c', r'CC(C)C(S)=O', 3)
RECENT_ELONGATION = BondDefiner('recent_elongation', r'O=C(C)CC(=O)S', 0, 1)
RECENT_ELONGATION_CC = BondDefiner('recent_elongation_CC', r'O=C(C)CC(=O)S', 1, 2)
RECENT_REDUCTION_COH = BondDefiner('recent_reduction_C-OH', r'OC(C)CC(=O)S', 0, 1)
RECENT_REDUCTION_MMAL_CHIRAL_C = GroupDefiner('recent_reduction_mmal_chiral_c', r'CCC(C(=O)S)C', 2)
RECENT_REDUCTION_C = GroupDefiner('recent_reduction_mal', r'OCCC(=O)S', 2)
RECENT_REDUCTION_CC = BondDefiner('recent_reduction_C-C', r'OCCC(=O)S', 1, 2)
RECENT_REDUCTION_CC_SHIFTED = BondDefiner('recent_reduction_C-C_shifted', r'CC(O)CC(S)=O', 0, 1)
RECENT_DEHYDRATION = BondDefiner('recent_dehydration', r'SC(C=CC)=O', 2, 3)
RECENT_EONYL_REDUCTION = BondDefiner('recent_eonyl_reduction', r"CCCC(S)=O", 2, 3)
RECENT_EONYL_REDUCTION_CC = BondDefiner('recent_eonyl_reduction', r"CCCC(S)=O", 2, 1)
RECENT_REDUCTION_OH = BondDefiner('recent_reduction_C-H', r'[H]OC(C)CC(S)=O', 0, 1)
RECENT_REDUCTION_TOP_C = GroupDefiner('recent_reduction_top_c', r'[H]C(C)=C([H])C(S)=O', 5)
RECENT_REDUCTION_TOP_H = GroupDefiner('recent_reduction_top_h', r'[H]C(C)=C([H])C(S)=O', 4)
RECENT_REDUCTION_TOP_METHYL = GroupDefiner('recent_reduction_top_methyl', r'[H]\C(C)=C(\C)C(S)=O', 4)
RECENT_REDUCTION_BOTTOM_C = GroupDefiner('recent_reduction_bottom_c', r'[H]C(C)=C([H])C(S)=O', 2)
RECENT_REDUCTION_BOTTOM_H = GroupDefiner('recent_reduction_bottom_h', r'[H]C(C)=C([H])C(S)=O', 0)
RECENT_REDUCTION_SHIFTED_TOP_C = GroupDefiner('recent_reduction_shifted_top_c', r'[H]\C(C)=C(\[H])CC(S)=O', 5)
RECENT_REDUCTION_SHIFTED_TOP_H = GroupDefiner('recent_reduction_shifted_top_h', r'[H]\C(C)=C(\[H])CC(S)=O', 4)
RECENT_REDUCTION_SHIFTED_BOTTOM_C = GroupDefiner('recent_reduction_shifted_bottom_c', r'[H]\C(C)=C(\[H])CC(S)=O', 2)
RECENT_REDUCTION_SHIFTED_BOTTOM_H = GroupDefiner('recent_reduction_shifted_bottom_h', r'[H]\C(C)=C(\[H])CC(S)=O', 0)
RECENT_REDUCTION_SHIFTED_BOTTOM_METHYL = GroupDefiner('recent_reduction_shifted_bottom_methyl',
                                                      r'[H]\C(CC(S)=O)=C(\C)CC', 7)
S_KR = GroupDefiner('C1 atom before KR reaction', r'SC(C)=O', 0)
ER_MMAL_CARBON = GroupDefiner('Chiral carbon atom after enoylreduction of mmal', r'SC(=O)C(C)CC', 3)
ER_S_CARBON = GroupDefiner('S-carbon atom after enoylreduction of mmal', r'SC(=O)C(C)CC', 1)
HYDROXYL_GROUP_TWO_MODULES_UPSTREAM_ALPHA_WITH_DOUBLE_BOND = BondDefiner("hydroxyl_group_two_module_upstream_alpha_with_double_bond",
                                                                         r"OC\C=C\CCC(S)=O", 0, 1)
HYDROXYL_GROUP_TWO_MODULES_UPSTREAM_BETA_WITH_DOUBLE_BOND = BondDefiner("hydroxyl_group_two_module_upstream_beta_with_double_bond",
                                                                        r"OCC\C=C\CCC(S)=O", 0, 1)
HYDROXYL_GROUP_TWO_MODULES_UPSTREAM_ALPHA_WITH_DOUBLE_BOND_SHIFTED = BondDefiner("hydroxyl_group_two_module_upstream_alpha_with_double_bond_shifted",
                                                                                 r"O\C=C\CCCC(S)=O", 0, 1)
HYDROXYL_GROUP_TWO_MODULES_UPSTREAM_BETA_WITH_DOUBLE_BOND_SHIFTED = BondDefiner("hydroxyl_group_two_module_upstream_beta_with_double_bond_shifted",
                                                                                r"OC\C=C\CCCC(S)=O", 0, 1)
HYDROXYL_GROUP_TWO_MODULES_UPSTREAM_ALPHA = BondDefiner("hydroxyl_group_two_module_upstream_alpha",
                                                        "OCCCCCC(S)=O", 0, 1)
HYDROXYL_GROUP_TWO_MODULES_UPSTREAM_BETA = BondDefiner("hydroxyl_group_two_module_upstream_beta",
                                                       "OCCCCCCC(S)=O", 0, 1)


# STILL MISSING: E/Z-configured double bonds, E/Z-Gamma-beta-dehydrogenase

def ketoreduction(chain_intermediate: Structure, kr_type: KRDomainSubtype) -> Tuple[Structure, bool]:
    """
    Performs the ketoreductase reaction on the PKS chain intermediate, returns
    the reaction product as a PIKAChU Structure object

    chain_intermediate: PIKAChU Structure object, PKS chain intermediate just
    after a PKS elongation step
    """

    if kr_type.name == 'C1':
        return chain_intermediate, False

    chiral_c = None

    # Find chiral carbon atom
    chiral_c_locations = find_atoms(RECENT_REDUCTION_MMAL_CHIRAL_C, chain_intermediate)

    # Identify beta-ketone bond, identify O- and C-atom participating in bond

    beta_ketone_bonds = find_bonds(RECENT_ELONGATION, chain_intermediate)

    if not len(beta_ketone_bonds) == 1:
        return chain_intermediate, False

    beta_ketone_bond = beta_ketone_bonds[0]

    carbonyl_oxygen = beta_ketone_bond.get_neighbour('O')
    carbonyl_carbon = beta_ketone_bond.get_neighbour('C')

    if not carbonyl_carbon or not carbonyl_oxygen:
        raise ValueError("Beta-ketone bond does not have carbon or oxygen neighbour.")

    # Identify beta-ketone bond, identify O- and C-atom participating in bond
    if kr_type.name != "C2":

        # Change carbonyl bond to single bond
        beta_ketone_bond.make_single()

        # Add H atom to form hydroxyl group and another H to the C
        for neighbour in beta_ketone_bond.neighbours:
            chain_intermediate.add_atom('H', [neighbour])

        # Add atom annotations to new hydrogen atoms
        initialise_atom_attributes(chain_intermediate)

        chain_intermediate.refresh_structure()
        carbonyl_carbon = chain_intermediate.get_atom(carbonyl_carbon)

        h_atom = carbonyl_carbon.get_neighbour('H')
        o_atom = carbonyl_carbon.get_neighbour('O')
        c_atoms = carbonyl_carbon.get_neighbours('C')

        # Define the two carbons adjacent to the carbonyl atom as 'top_c' and 'bottom_c'

        top_cs = find_atoms(RECENT_REDUCTION_C, chain_intermediate)
        assert len(top_cs) == 1

        top_c = top_cs[0]
        assert top_c in c_atoms

        bottom_c = None
        for atom in c_atoms:
            if atom != top_c:
                bottom_c = atom

        assert bottom_c

        # Set the chirality of the carbonyl carbon

        counterclockwise_order = [h_atom, o_atom, top_c, bottom_c]

        if kr_type.name.startswith('A'):

            if same_chirality(counterclockwise_order, carbonyl_carbon.neighbours):
                carbonyl_carbon.chiral = 'clockwise'
            else:
                carbonyl_carbon.chiral = 'counterclockwise'

        elif kr_type.name.startswith('B'):
            if same_chirality(counterclockwise_order, carbonyl_carbon.neighbours):
                carbonyl_carbon.chiral = 'counterclockwise'
            else:
                carbonyl_carbon.chiral = 'clockwise'

        # If the type of KR domain is not known, the chirality of the carbonyl carbon is set to None
        elif kr_type.name == "UNKNOWN":
            carbonyl_carbon.chiral = None

        else:
            raise ValueError(f'KR domain of type {kr_type.name} is not supported by RAIChU or does not exist')

    # See if the previous elongation step was performed using methylmalonyl-CoA,
    # perform epimerization if required

    if chiral_c_locations:

        chiral_c = chiral_c_locations[0]

    if chiral_c:
        sulphur_locations = find_atoms(S_KR, chain_intermediate)
        assert len(sulphur_locations) == 1
        sulphur = sulphur_locations[0]
        assert sulphur

        c_1 = None
        for neighbour in sulphur.neighbours:
            if neighbour.annotations.in_central_chain and neighbour.type == 'C':
                c_1 = neighbour
                break
        assert c_1

        c_3 = None
        for neighbour in chiral_c.neighbours:
            if neighbour.annotations.in_central_chain and neighbour != c_1:
                c_3 = neighbour
        assert c_3

        hydrogen = None
        first_sidechain_atom = None

        for atom in chiral_c.neighbours:
            if atom != c_1 and atom != c_3 and atom.type != 'H' and \
                    not atom.annotations.in_central_chain:
                first_sidechain_atom = atom
            elif atom.type == 'H':
                hydrogen = atom

        assert hydrogen
        assert first_sidechain_atom

        clockwise_order = [first_sidechain_atom, c_1, c_3, hydrogen]

        if kr_type.name.endswith('1'):
            if same_chirality(clockwise_order, chiral_c.neighbours):
                chiral_c.chiral = 'clockwise'
            else:
                chiral_c.chiral = 'counterclockwise'
        elif kr_type.name.endswith('2'):
            if same_chirality(clockwise_order, chiral_c.neighbours):
                chiral_c.chiral = 'counterclockwise'
            else:
                chiral_c.chiral = 'clockwise'
        elif kr_type.name == "UNKNOWN":
            chiral_c.chiral = None
        else:
            raise ValueError(f'KR domain of type {kr_type.name} is not supported by RAIChU or does not exist')

    chain_intermediate.refresh_structure()
    return chain_intermediate, True


def dehydration(chain_intermediate: Structure, chirality=None) -> Tuple[Structure, bool]:
    """
    Performs the dehydratase reaction on the PKS chain intermediate, returns
    the reaction product as a PIKAChU Structure object

    chain_intermediate: PIKAChU Structure object, PKS chain intermediate where
    the beta ketone group has been recently reduced by the KR domain
    """

    # Find and define the atoms that participate in the bond changes
    co_bonds = find_bonds(RECENT_REDUCTION_COH, chain_intermediate)
    cc_bonds = find_bonds(RECENT_REDUCTION_CC, chain_intermediate)

    if not len(co_bonds) == 1 or not len(cc_bonds) == 1:
        return chain_intermediate, False

    c1 = None
    c2 = None
    o_oh = None

    co_bond = co_bonds[0]

    for neighbour in co_bond.neighbours:
        if neighbour.type == 'C':
            c2 = neighbour
        elif neighbour.type == 'O':
            o_oh = neighbour

    cc_bond = cc_bonds[0]

    for neighbour in cc_bond.neighbours:
        if neighbour != c2:
            c1 = neighbour

    # Remove hydroxyl group from c2
    chain_intermediate.break_bond(co_bond)

    # Remove H-atom from c1
    bond_to_break = None
    hydrogen = None

    for atom in chain_intermediate.graph:
        if atom == c1:
            for neighbour in atom.neighbours:
                if neighbour.type == 'H':
                    hydrogen = neighbour
                    for bond in neighbour.bonds:
                        bond_to_break = bond
                    break

    assert bond_to_break and hydrogen

    chain_intermediate.break_bond(bond_to_break)

    # Patch. When the bond is broken, the electron is not removed from
    # the orbitals of the C atom.
    c1 = chain_intermediate.get_atom(c1)

    # Make double c1=c2 bond
    double_bond = chain_intermediate.bond_lookup[c1][c2]
    double_bond.make_double()
    # Make bond between removed hydroxyl group and hydrogen atom to form water
    next_bond_nr = chain_intermediate.find_next_bond_nr()
    chain_intermediate.make_bond(hydrogen, o_oh, next_bond_nr)

    # Remove separate water structure from the Structure object
    structure_1, structure_2 = chain_intermediate.split_disconnected_structures()
    if c1 in structure_1.graph:
        chain_intermediate = structure_1
    else:
        chain_intermediate = structure_2

    chain_intermediate.refresh_structure()
    double_bond = chain_intermediate.bond_lookup[c1][c2]
    # implement stereochemistry
    if chirality:
        main_chain_top_c = find_atoms(RECENT_REDUCTION_TOP_C, chain_intermediate)[0]
        main_chain_bottom_c = find_atoms(RECENT_REDUCTION_BOTTOM_C, chain_intermediate)[0]
        main_chain_top_h = find_atoms(RECENT_REDUCTION_TOP_H, chain_intermediate)
        if not main_chain_top_h:
            main_chain_top_h = find_atoms(RECENT_REDUCTION_TOP_METHYL, chain_intermediate)[0]
        else:
            main_chain_top_h = main_chain_top_h[0]
        main_chain_bottom_h = find_atoms(RECENT_REDUCTION_BOTTOM_H, chain_intermediate)[0]
        if chirality == "E":
            double_bond.chiral_dict = {main_chain_top_c: {main_chain_bottom_c: 'trans',
                                                          main_chain_bottom_h: 'cis'},
                                       main_chain_top_h: {main_chain_bottom_c: 'cis',
                                                          main_chain_bottom_h: 'trans'},
                                       main_chain_bottom_c: {main_chain_top_c: 'trans',
                                                             main_chain_top_h: 'cis'},
                                       main_chain_bottom_h: {main_chain_top_c: 'cis',
                                                             main_chain_top_h: 'trans'}}
        if chirality == "Z":
            double_bond.chiral_dict = {main_chain_top_c: {main_chain_bottom_c: 'cis',
                                                          main_chain_bottom_h: 'trans'},
                                       main_chain_top_h: {main_chain_bottom_c: 'trans',
                                                          main_chain_bottom_h: 'cis'},
                                       main_chain_bottom_c: {main_chain_top_c: 'cis',
                                                             main_chain_top_h: 'trans'},
                                       main_chain_bottom_h: {main_chain_top_c: 'trans',
                                                             main_chain_top_h: 'cis'}}
        double_bond.chiral = True
    chain_intermediate.refresh_structure()

    return chain_intermediate, True


def enoylreduction(chain_intermediate: Structure, er_subtype: ERDomainSubtype) -> Tuple[Structure, bool]:
    """
    Performs the enoylreductase reaction on the PKS chain intermediate, returns
    the reaction product as a PIKAChU Structure object

    chain_intermediate: PIKAChU Structure object, PKS chain intermediate where
    the beta ketone group has been recently reduced and dehydrated by the KR
    and ER domains, respectively
    er_subtype: ERDomainSubtype
    """

    for atom in chain_intermediate.graph:
        atom.set_connectivity()

    # Find double bond, change to single bond
    double_cc_bonds = find_bonds(RECENT_DEHYDRATION, chain_intermediate)
    if not len(double_cc_bonds) == 1:
        return chain_intermediate, False

    double_cc_bond = double_cc_bonds[0]
    double_cc_bond.make_single()

    # Add H-atom to the C atoms participating in the new single bond
    for neighbour in double_cc_bond.neighbours:
        chain_intermediate.add_atom('H', [neighbour])

    # Give annotation to added H-atom
    initialise_atom_attributes(chain_intermediate)
    chain_intermediate.refresh_structure()

    er_carbons = set(find_atoms(ER_MMAL_CARBON, chain_intermediate))
    if len(er_carbons) == 1 and er_subtype.name != "UNKNOWN":
        er_carbon = list(er_carbons)[0]
        if len(er_carbon.get_neighbours('H')) == 1 and len(er_carbon.get_neighbours('C')) == 3:
            s_carbons = set(find_atoms(ER_S_CARBON, chain_intermediate))
            assert len(s_carbons) == 1

            s_carbon = list(s_carbons)[0]
            bottom_carbon = None
            sidechain_carbon = None
            for carbon_neighbour in er_carbon.get_neighbours('C'):
                if carbon_neighbour.annotations.in_central_chain and carbon_neighbour != s_carbon:
                    bottom_carbon = carbon_neighbour
                elif not carbon_neighbour.annotations.in_central_chain:
                    sidechain_carbon = carbon_neighbour

            hydrogen_neighbour = er_carbon.get_neighbour('H')
            assert bottom_carbon and sidechain_carbon and hydrogen_neighbour

            clockwise_s_order = [s_carbon, hydrogen_neighbour, sidechain_carbon, bottom_carbon]
            has_s_chirality = same_chirality(clockwise_s_order, er_carbon.neighbours)
            if er_subtype.name == 'S' and has_s_chirality:
                er_carbon.chiral = 'clockwise'
            elif er_subtype.name == 'S' and not has_s_chirality:
                er_carbon.chiral = 'counterclockwise'
            elif er_subtype.name == 'R' and not has_s_chirality:
                er_carbon.chiral = 'clockwise'
            elif er_subtype.name == 'R' and has_s_chirality:
                er_carbon.chiral = 'counterclockwise'
            else:
                raise ValueError(f"RAIChU does not support ER domain subtype {er_subtype.name}")

    return chain_intermediate, True


# trans-AT-PKS tailoring reactions
def hydroxylation(target_atom, structure: Structure) -> Structure:
    """
    Returns the hydroxylated structure thats hydroxylated at the target atom.

    structure: PIKAChU Structure object
    target_atom:  PIKAChU atom object
    """

    hydroxyl_group = read_smiles('o')
    hydroxyl_group.add_attributes(ATTRIBUTES, boolean=True)
    oxygen = hydroxyl_group.atoms[0]
    hydrogen_1 = hydroxyl_group.atoms[1]
    bond_1 = oxygen.get_bond(hydrogen_1)
    hydrogen_2 = target_atom.get_neighbour('H')
    if not hydrogen_2:
        raise Exception("Can't oxidate this atom!")

    bond_2 = target_atom.get_bond(hydrogen_2)

    hydroxylated_structure = combine_structures([structure, hydroxyl_group])

    hydroxylated_structure.break_bond(bond_1)
    hydroxylated_structure.break_bond(bond_2)

    hydroxylated_structure.make_bond(oxygen, target_atom, hydroxylated_structure.find_next_bond_nr())
    hydroxylated_structure.make_bond(hydrogen_1, hydrogen_2, hydroxylated_structure.find_next_bond_nr())

    structures = hydroxylated_structure.split_disconnected_structures()

    for s in structures:
        if oxygen.nr in s.atoms:
            return s


def methylation(target_atom, structure: Structure) -> Structure:
    """
    Returns the structure thats methylated at the target atom.

    structure: PIKAChU Structure object
    target_atom:  PIKAChU atom object
    """

    methyl_group = read_smiles('C')
    methyl_group.add_attributes(ATTRIBUTES, boolean=True)
    carbon = methyl_group.atoms[0]
    hydrogen_1 = methyl_group.atoms[1]
    bond_1 = carbon.get_bond(hydrogen_1)

    hydrogen_2 = target_atom.get_neighbour('H')
    if not hydrogen_2:
        raise Exception("Can't methylate this atom!")

    bond_2 = target_atom.get_bond(hydrogen_2)

    methylated_structure = combine_structures([structure, methyl_group])

    methylated_structure.break_bond(bond_1)
    methylated_structure.break_bond(bond_2)

    methylated_structure.make_bond(carbon, target_atom, methylated_structure.find_next_bond_nr())
    methylated_structure.make_bond(hydrogen_1, hydrogen_2, methylated_structure.find_next_bond_nr())

    structures = methylated_structure.split_disconnected_structures()

    for s in structures:
        if carbon.nr in s.atoms:
            return s


def find_OH_two_modules_upstream(structure: Structure):
    """
    Returns the O atom that is connected to the first hydroxygroup two modules upstream (also if chain contains double bonds)
    structure: PIKAChU Structure object
    """
    bonds = find_bonds(HYDROXYL_GROUP_TWO_MODULES_UPSTREAM_ALPHA, structure)
    if not bonds:
        bonds = find_bonds(HYDROXYL_GROUP_TWO_MODULES_UPSTREAM_BETA, structure)
    if not bonds:
        bonds = find_bonds(HYDROXYL_GROUP_TWO_MODULES_UPSTREAM_BETA_WITH_DOUBLE_BOND, structure)
    if not bonds:
        bonds = find_bonds(HYDROXYL_GROUP_TWO_MODULES_UPSTREAM_BETA_WITH_DOUBLE_BOND_SHIFTED, structure)
    return bonds


def smallest_cyclisation(structure: Structure) -> Tuple[Structure, bool]:
    """
    Cyclises with the OH group two modules upstream to the smalles possible ring

    structure: PIKAChU Structure object
    """
    # reduce Ketogroup first
    structure_oh, did_reduction = ketoreduction(structure, KRDomainSubtype(1))
    if not did_reduction:
        return structure, False

    oh_bond = find_OH_two_modules_upstream(structure_oh)
    if not oh_bond:
        print("No hydroxy group two modules upstream availiable")
        return structure, False

    oh_bond = oh_bond[0]
    h_bond = find_bonds(RECENT_REDUCTION_OH, structure_oh)[0]
    structure_oh = internal_condensation(structure_oh, oh_bond, h_bond)[0]

    return structure_oh, True


def alpha_methyl_transferase(structure: Structure) -> Tuple[Structure, bool]:
    """
    Returns the structure thats methylated at the alpha-c.

    structure: PIKAChU Structure object
    target_atom:  PIKAChU atom object
    """
    # find atom to add methylgroup
    alpha_c = find_atoms(RECENT_ALPHA_C, structure)
    if alpha_c:
        if alpha_c[0].has_neighbour("H"):
            structure = methylation(alpha_c[0], structure)
            return structure, True
        else:
            return structure, False
    else:
        return structure, False


def beta_methyl_transferase(structure: Structure) -> Tuple[Structure, bool]:
    """
    Returns the structure thats methylated at the beta-c.

    structure: PIKAChU Structure object
    target_atom:  PIKAChU atom object
    """
    # find atom to add methylgroup
    beta_c = find_atoms(RECENT_BETA_C, structure)
    if beta_c:
        if beta_c[0].has_neighbour("H"):
            structure = methylation(beta_c[0], structure)
            return structure, True
        else:
            return structure, False
    else:
        return structure, False


def beta_hydroxy_methyl_transferase(structure: Structure) -> Tuple[Structure, bool]:
    """
    Returns the structure thats methylated at the beta-oh-goup.

    structure: PIKAChU Structure object
    target_atom:  PIKAChU atom object
    """
    # find atom to add methylgroup

    beta_c_oh = find_atoms(RECENT_BETA_C_OH, structure)
    if beta_c_oh:
        if beta_c_oh[0].has_neighbour("H"):
            structure = methylation(beta_c_oh[0], structure)
            return structure, True
        else:
            return structure, False
    else:
        return structure, False


def alpha_hydroxylase(structure: Structure) -> Tuple[Structure, bool]:
    """
    Returns the structure thats hydroxylated at the alpha-c.

    structure: PIKAChU Structure object
    target_atom:  PIKAChU atom object
    """
    # find atom to add methylgroup
    alpha_c = find_atoms(RECENT_ALPHA_C, structure)
    if alpha_c:
        if alpha_c[0].has_neighbour("H"):
            structure = hydroxylation(alpha_c[0], structure)
            return structure, True
        else:
            return structure, False
    else:
        return structure, False


def alpha_L_methyl_transferase(structure: Structure) -> Tuple[Structure, bool]:
    """
    Returns the structure thats methylated at the alpha- c in L configuration.

    structure: PIKAChU Structure object
    target_atom:  PIKAChU atom object
    """
    # find atom to add methylgroup
    alpha_c = find_atoms(RECENT_ALPHA_C, structure)
    if alpha_c:
        if alpha_c[0].has_neighbour("H"):
            structure = methylation(alpha_c[0], structure)
            alpha_c = alpha_c[0]
            structure.refresh_structure()
            h_atom = alpha_c.get_neighbour('H')
            c_atoms = alpha_c.get_neighbours('C')

            # Define the two carbons adjacent to the carbonyl atom as 'top_c' and 'bottom_c'
            methyl_c = find_atoms(RECENT_ALPHA_METHYL_C, structure)[0]
            if not methyl_c:
                return structure, False
            top_cs = find_atoms(FIRST_C, structure)

            if not len(top_cs) == 1:
                return structure, False
            top_c = top_cs[0]
            if top_c not in c_atoms:
                return structure, False

            bottom_c = None
            for atom in c_atoms:
                if atom != top_c and atom != methyl_c:
                    bottom_c = atom

            if not bottom_c:
                return structure, False

            # Set the chirality of the carbonyl carbon

            counterclockwise_order = [h_atom, methyl_c, bottom_c, top_c]

            if same_chirality(counterclockwise_order, alpha_c.neighbours):
                alpha_c.chiral = 'clockwise'
            else:
                alpha_c.chiral = 'counterclockwise'
            structure.refresh_structure()
            return structure, True
        else:
            return structure, False
    else:
        return structure, False


def gamma_beta_dehydratase(chain_intermediate: Structure, chirality=None) -> Tuple[Structure, bool]:
    """
    Performs the dehydratase reaction on the PKS chain intermediate wuth shifted double bonds, returns
    the reaction product as a PIKAChU Structure object

    chain_intermediate: PIKAChU Structure object, PKS chain intermediate where
    the beta ketone group has been recently reduced by the KR domain
    subtype: E or Z, depending on chirality, if none, no chirality specified
    """
    # Find and define the atoms that participate in the bond changes
    co_bonds = find_bonds(RECENT_REDUCTION_COH, chain_intermediate)
    cc_bonds = find_bonds(RECENT_REDUCTION_CC_SHIFTED, chain_intermediate)

    if not len(co_bonds) == 1 or not len(cc_bonds) == 1:
        return chain_intermediate, False

    c1 = None
    c2 = None
    o_oh = None

    co_bond = co_bonds[0]

    for neighbour in co_bond.neighbours:
        if neighbour.type == 'C':
            c2 = neighbour
        elif neighbour.type == 'O':
            o_oh = neighbour

    cc_bond = cc_bonds[0]

    for neighbour in cc_bond.neighbours:
        if neighbour != c2:
            c1 = neighbour

    # Remove hydroxyl group from c2
    chain_intermediate.break_bond(co_bond)

    # Remove H-atom from c1
    bond_to_break = None
    hydrogen = None

    for atom in chain_intermediate.graph:
        if atom == c1:
            for neighbour in atom.neighbours:
                if neighbour.type == 'H':
                    hydrogen = neighbour
                    for bond in neighbour.bonds:
                        bond_to_break = bond
                    break

    assert bond_to_break and hydrogen

    chain_intermediate.break_bond(bond_to_break)

    # Patch. When the bond is broken, the electron is not removed from
    # the orbitals of the C atom.
    c1 = chain_intermediate.get_atom(c1)

    # Make double c1=c2 bond
    double_bond = chain_intermediate.bond_lookup[c1][c2]
    double_bond.make_double()

    # Make bond between removed hydroxyl group and hydrogen atom to form water
    next_bond_nr = chain_intermediate.find_next_bond_nr()
    chain_intermediate.make_bond(hydrogen, o_oh, next_bond_nr)

    # Remove separate water structure from the Structure object
    structure_1, structure_2 = chain_intermediate.split_disconnected_structures()
    if c1 in structure_1.graph:
        chain_intermediate = structure_1
    else:
        chain_intermediate = structure_2

    chain_intermediate.refresh_structure()

    # implement stereochemistry
    if chirality:

        main_chain_top_c = find_atoms(RECENT_REDUCTION_SHIFTED_TOP_C, chain_intermediate)[0]
        main_chain_bottom_c = find_atoms(RECENT_REDUCTION_SHIFTED_BOTTOM_C, chain_intermediate)[0]
        main_chain_top_h = find_atoms(RECENT_REDUCTION_SHIFTED_TOP_H, chain_intermediate)
        main_chain_bottom_h = find_atoms(RECENT_REDUCTION_SHIFTED_BOTTOM_H, chain_intermediate)[0]
        if not main_chain_bottom_h:
            main_chain_bottom_h = find_atoms(RECENT_REDUCTION_SHIFTED_BOTTOM_METHYL, chain_intermediate)[0]
        else:
            main_chain_bottom_h = main_chain_bottom_h[0]
        if chirality == "E":
            double_bond.chiral_dict = {main_chain_top_c: {main_chain_bottom_c: 'trans',
                                                          main_chain_bottom_h: 'cis'},
                                       main_chain_top_h: {main_chain_bottom_c: 'cis',
                                                          main_chain_bottom_h: 'trans'},
                                       main_chain_bottom_c: {main_chain_top_c: 'trans',
                                                             main_chain_top_h: 'cis'},
                                       main_chain_bottom_h: {main_chain_top_c: 'cis',
                                                             main_chain_top_h: 'trans'}}
        if chirality == "Z":
            double_bond.chiral_dict = {main_chain_top_c: {main_chain_bottom_c: 'trans',
                                                          main_chain_bottom_h: 'cis'},
                                       main_chain_top_h: {main_chain_bottom_c: 'cis',
                                                          main_chain_bottom_h: 'trans'},
                                       main_chain_bottom_c: {main_chain_top_c: 'cis',
                                                             main_chain_top_h: 'trans'},
                                       main_chain_bottom_h: {main_chain_top_c: 'trans',
                                                             main_chain_top_h: 'cis'}}
        double_bond.chiral = True
    chain_intermediate.refresh_structure()
    return chain_intermediate, True



import pygame
from interactive.style import BLACK, MODULE_PADDING, GENE_LABEL_SIZE


class InsertionPoint:
    def __init__(self, gene, mouse):
        self.gene = gene
        self.insertion_point, self.insertion_x = self.get_insertion_point(mouse)

    def get_insertion_point(self, mouse):

        insertion_point = 0
        insertion_x = self.gene.x + GENE_LABEL_SIZE + MODULE_PADDING / 2

        for i, module in enumerate(self.gene.modules):
            if module.x < mouse[0]:
                insertion_point = i + 1
                insertion_x = module.x + module.width + MODULE_PADDING / 2
            else:
                break

        return insertion_point, insertion_x

    def draw(self, screen):
        length = self.gene.height
        y_start = self.gene.y
        y_end = y_start + length
        pygame.draw.line(screen, BLACK, (self.insertion_x, y_start), (self.insertion_x, y_end - 1))



from typing import List, Union

from raichu.cluster import Cluster
from raichu.domain.domain import TailoringDomain, CarrierDomain, SynthesisDomain, RecognitionDomain, \
    TerminationDomain, UnknownDomain, Domain
from raichu.module import PKSModuleSubtype, NRPSModule, LinearPKSModule, IterativePKSModule, TransATPKSModule,\
    ModuleType
from raichu.domain.domain_types import TailoringDomainType, TerminationDomainType, CarrierDomainType, \
    SynthesisDomainType, RecognitionDomainType
from dataclasses import dataclass


DOMAIN_TO_SUPERTYPE = {}
for domain_name in TailoringDomainType.__members__:
    DOMAIN_TO_SUPERTYPE[domain_name] = TailoringDomain
for domain_name in CarrierDomainType.__members__:
    DOMAIN_TO_SUPERTYPE[domain_name] = CarrierDomain
for domain_name in SynthesisDomainType.__members__:
    DOMAIN_TO_SUPERTYPE[domain_name] = SynthesisDomain
for domain_name in RecognitionDomainType.__members__:
    DOMAIN_TO_SUPERTYPE[domain_name] = RecognitionDomain
for domain_name in TerminationDomainType.__members__:
    DOMAIN_TO_SUPERTYPE[domain_name] = TerminationDomain
DOMAIN_TO_SUPERTYPE["UNKNOWN"] = UnknownDomain


@dataclass
class DomainRepresentation:
    gene_name: Union[str, None]
    type: str
    subtype: Union[str, None]
    name: Union[str, None]
    active: bool
    used: bool


@dataclass
class ModuleRepresentation:
    type: str
    subtype: Union[str, None]
    substrate: str
    domains: List[DomainRepresentation]


@dataclass
class ClusterRepresentation:
    modules: List[ModuleRepresentation]


def make_domain(domain_repr: DomainRepresentation, substrate: str, strict: bool = True) -> Domain:
    domain_class = DOMAIN_TO_SUPERTYPE.get(domain_repr.type)
    if domain_class:
        if domain_class == RecognitionDomain:
            domain = domain_class(domain_repr.type, substrate, domain_subtype=domain_repr.subtype,
                                  active=domain_repr.active,
                                  used=domain_repr.used)
        elif domain_class == UnknownDomain:
            domain = UnknownDomain(domain_repr.name)
        else:
            domain = domain_class(domain_repr.type, domain_subtype=domain_repr.subtype, active=domain_repr.active,
                                  used=domain_repr.used)
    elif strict:
        raise ValueError(f"Unrecognised domain type: {domain_repr.type}")
    else:
        domain = UnknownDomain(domain_repr.name)

    return domain


def build_cluster(cluster_repr: ClusterRepresentation, strict: bool = True) -> Cluster:

    genes = set()

    modules = []
    previous_domain = None
    for i, module_repr in enumerate(cluster_repr.modules):
        if i == 0:
            starter = True
        else:
            starter = False

        if i == len(cluster_repr.modules) - 1:
            terminator = True
        else:
            terminator = False

        domains = []
        for domain_repr in module_repr.domains:
            domain = make_domain(domain_repr, module_repr.substrate,
                                 strict=strict)

            if domain_repr.gene_name is not None:
                if previous_domain:
                    previous_gene = previous_domain.gene
                    if previous_gene != domain_repr.gene_name and domain_repr.gene_name in genes:
                        raise ValueError(f"Gene name '{previous_gene}' already assigned to upstream domain(s).")

                domain.set_gene(domain_repr.gene_name)
                genes.add(domain_repr.gene_name)
            domains.append(domain)
            previous_domain = domain

        module_type = ModuleType.from_string(module_repr.type)
        if module_type.name == 'PKS':
            if module_repr.subtype is not None:
                module_subtype = PKSModuleSubtype.from_string(module_repr.subtype)

                if module_subtype.name == 'PKS_CIS':
                    module = LinearPKSModule(i, domains, starter=starter, terminator=terminator)
                elif module_subtype.name == 'PKS_TRANS':
                    module = TransATPKSModule(i, domains, starter=starter, terminator=terminator)
                elif module_subtype.name == 'PKS_ITER':
                    module = IterativePKSModule(i, domains, starter=starter, terminator=terminator)
                else:
                    raise ValueError(f"Unrecognised PKS module subtype: {module_subtype}.")
            else:
                raise ValueError("PKS module subtype must be specified.")

        elif module_type.name == 'NRPS':
            if module_repr.subtype is None:
                module = NRPSModule(i, domains, starter=starter, terminator=terminator)
            else:
                raise ValueError("NRPS module subtypes are currently not supported. Please pass None.")
        else:
            raise ValueError(f"Unrecognised module type: {module_repr.type}")

        modules.append(module)

    cluster = Cluster(modules)

    # return cluster_with_processed_trans_at_pks

    return cluster

def draw_cluster(cluster_repr: ClusterRepresentation) -> None:
    cluster = build_cluster(cluster_repr)
    cluster.compute_structures(compute_cyclic_products=False)
    cluster.draw_cluster()
def get_spaghettis(cluster_repr: ClusterRepresentation) -> List[str]:

    cluster = build_cluster(cluster_repr)
    # for module in cluster.modules:
    #     print(module.domains)
    cluster.compute_structures(compute_cyclic_products=False)
    cluster.draw_cluster()
    spaghettis = cluster.draw_spaghettis()

    return spaghettis


if __name__ == "__main__":
    cluster_repr = ClusterRepresentation([ModuleRepresentation("PKS", "PKS_CIS", "ACETYL_COA",
                                                               [DomainRepresentation("gene 1", 'AT', None, None, True,
                                                                                     True),
                                                                DomainRepresentation("gene 1", 'KR', 'A1', None, True,
                                                                                     True),
                                                                DomainRepresentation("gene 1", 'DH', None, None, True,
                                                                                     True),
                                                                DomainRepresentation("gene 1", 'ACP', None, None, True,
                                                                                     True)
                                                                ]),
                                          ModuleRepresentation("PKS", "PKS_CIS", "METHYLMALONYL_COA",
                                                               [DomainRepresentation("gene 1", 'KS',
                                                                                     None, None, True,
                                                                                     True),
                                                                DomainRepresentation("gene 1", 'KS',
                                                                                     None, None, True,
                                                                                     False),
                                                                DomainRepresentation("gene 1", 'AT', None, None, True,
                                                                                     True),
                                                                DomainRepresentation("gene 1", 'AT', None, None, True,
                                                                                     False),
                                                                DomainRepresentation("gene 1", 'DH', None, None, True,
                                                                                     True),
                                                                DomainRepresentation("gene 1", 'ER', None, None, True,
                                                                                     True),
                                                                DomainRepresentation("gene 1", 'ACP', None, None, True,
                                                                                     True)
                                                                ]),
                                          ModuleRepresentation("NRPS", None, "glycine",
                                                               [DomainRepresentation("gene 1", 'C', None, None, True,
                                                                                     True),
                                                                DomainRepresentation("gene 1", 'nMT', None, None, True,
                                                                                     True),
                                                                DomainRepresentation("gene 1", 'nMT', None, None, True,
                                                                                     False),
                                                                DomainRepresentation("gene 1", 'A', None, None, True,
                                                                                     True),
                                                                DomainRepresentation("gene 1", 'PCP', None, None, True,
                                                                                     True),
                                                                DomainRepresentation("gene 1", 'E', None, None, True,
                                                                                     True)
                                                                ]),
                                          ModuleRepresentation("PKS", "PKS_CIS", "METHYLMALONYL_COA",
                                                               [DomainRepresentation("gene 1", 'AT', None, None, True,
                                                                                     True),
                                                                DomainRepresentation("gene 2", 'KR', 'A1', None, True,
                                                                                     True),
                                                                DomainRepresentation("gene 2", 'DH', None, None, True,
                                                                                     True),
                                                                DomainRepresentation("gene 2", 'ACP', None, None, True,
                                                                                     True)
                                                                ]),
                                          ModuleRepresentation("PKS", "PKS_CIS", "METHYLMALONYL_COA",
                                                               [DomainRepresentation("gene 2", 'AT', None, None, True,
                                                                                     True),
                                                                DomainRepresentation("gene 3", 'KR', 'A1', None, True,
                                                                                     True),
                                                                DomainRepresentation("gene 3", 'DH', None, None, True,
                                                                                     True),
                                                                DomainRepresentation("gene 3", 'ACP', None, None, True,
                                                                                     True)
                                                                ]),
                                          ModuleRepresentation("NRPS", None, "tryptophan",
                                                               [DomainRepresentation("gene 4", 'C', None, None, True,
                                                                                     True),
                                                                DomainRepresentation("gene 4", 'C', None, None, True,
                                                                                     False),
                                                                DomainRepresentation("gene 4", 'A', None, None, True,
                                                                                     True),
                                                                DomainRepresentation("gene 4", 'A', None, None, True,
                                                                                     False),
                                                                DomainRepresentation("gene 4", 'PCP', None, None, True,
                                                                                     True),
                                                                DomainRepresentation("gene 4", 'PCP', None, None, True,
                                                                                     False)
                                                                ]),
                                          ModuleRepresentation("NRPS", None, "lysine",
                                                               [DomainRepresentation("gene 5", 'C', None, None, True,
                                                                                     True),
                                                                DomainRepresentation("gene 5", 'A', None, None, True,
                                                                                     True),
                                                                DomainRepresentation("gene 6", 'PCP', None, None, True,
                                                                                     True),
                                                                DomainRepresentation("gene 7", 'TE', None, None, True,
                                                                                     True)
                                                                ]),
                                          ModuleRepresentation("NRPS", None, "lysine",
                                                               [DomainRepresentation("gene 7", 'C', None, None, True,
                                                                                     True),
                                                                DomainRepresentation("gene 7", 'A', None, None, True,
                                                                                     True),
                                                                DomainRepresentation("gene 7", 'UNKNOWN', None, 'oMT',
                                                                                     True, True),
                                                                DomainRepresentation("gene 7", 'UNKNOWN', None, 'lalala',
                                                                                     True, True),

                                                                DomainRepresentation("gene 7", 'PCP', None, None, True,
                                                                                     True),
                                                                DomainRepresentation("gene 7", 'TE', None, None, True,
                                                                                     True)
                                                                ])
                                          ])

    draw_cluster(cluster_repr)



import os

from Bio import SeqIO
from interactive.gene import Gene

from interactive.substrate import Substrate
from interactive.parsers import parse_smiles
import interactive.flatfiles

FLATFILES = os.path.dirname(interactive.flatfiles.__file__)
PARAS_SMILES = os.path.join(FLATFILES, "PARAS_smiles.txt")


NRPS_NAME_TO_SMILES = parse_smiles(PARAS_SMILES)

PKS_NAME_TO_SMILES = {'malonylcoa': "CC(=O)S",
                      'methylmalonylcoa': "CCC(=O)S",
                      'methoxymalonylcoa': "COCC(=O)S",
                      'ethylmalonylcoa': "CCCC(=O)S",
                      'wildcard': "[*]CC(=O)S"}

NRPS_PKS_RULES = {'biosynthetic-additional (rule-based-clusters) AMP-binding',
                  'biosynthetic-additional (rule-based-clusters) PP-binding',
                  'biosynthetic (rule-based-clusters) NRPS: Condensation',
                  'biosynthetic (rule-based-clusters) T1PKS: PKS_AT',
                  'biosynthetic (rule-based-clusters) T1PKS: mod_KS'}

AS_TO_NRPS = {"ala": 'Alanine',
              "arg": "Arginine",
              "asn": "Asparagine",
              "asp": "AsparticAcid",
              "cys": "Cysteine",
              "gln": "Glutamine",
              "glu": "GlutamicAcid",
              "gly": "Glycine",
              "his": "Histidine",
              "ile": "Isoleucine",
              "leu": "Leucine",
              "lys": "Lysine",
              "met": "Methionine",
              "phe": "Phenylalanine",
              "pro": "Proline",
              "ser": "Serine",
              "thr": "Threonine",
              "trp": "Tryptophan",
              "tyr": "Tyrosine",
              "val": "Valine",
              "3-me-glu": "4-methylglutamicacid",
              "4ppro": "nrp",
              "aad": "2-Aminoadipicacid",
              'abu': "2-Aminobutyricacid",
              'aeo': '2-Amino-9,10-epoxy-8-oxodecanoidacid',
              'ala-b': 'beta-Alanine',
              'ala-d': 'D-Alanine',
              'allo-thr': "allo-Threonine",
              'b-ala': 'beta-Alanine',
              'beta-ala': 'beta-Alanine',
              'bmt': "4-butenyl-4-methylthreonine",
              'cap': "capreomycidine",
              'bht': 'nrp',
              'dab': "2,4-diaminobutyricacid",
              'dhb': "2,3-dihydroxybenzoicacid",
              'dhpg': "3,5-dihydroxyphenylglycine",
              'dht': "Dehydrobutyrine",
              'dpg': "3,5-dihydroxyphenylglycine",
              'hiv': "2-hydroxyisovalerate",
              'hiv-d': "D-2-hydroxyisovalerate",
              'hmp-d': 'nrp',
              'horn': 'nrp',
              'hpg': "HydroxyPhenylGlycine",
              'hyv': "4-Hydroxyvaline",
              'hyv-d': 'nrp',
              'iva': "isovaline",
              'lys-b': "beta-lysine",
              'orn': "Ornithine",
              'phg': "phenylglycine",
              'pip': "Pipecolicacid",
              'sal': "salicylicacid",
              'tcl': 'nrp',
              'vol': "valinol",
              'LDAP': 'nrp',
              'meval': "tert-Leu",
              'alle': "allo-Isoleucine",
              'alaninol': "Alaninol",
              'N-(1,1-dimethyl-1-allyl)Trp': 'nrp',
              'd-lyserg': "D-lysergicacid",
              'ser-thr': 'nrp',
              'mephe': 'nrp',
              'haorn': 'nrp',
              'hasn': 'nrp',
              'hforn': 'nrp',
              's-nmethoxy-trp': 'nrp',
              'alpha-hydroxy-isocaproic-acid': 'nrp',
              'MeHOval': 'nrp',
              '2-oxo-isovaleric-acid': "alpha-ketoisovalericacid",
              'aoda': 'nrp',
              'X': 'nrp'}

AS_TO_PKS = {'mal': 'malonylcoa',
             'mmal': 'methylmalonylcoa',
             'mxmal': 'methoxymalonylcoa',
             'emal': 'ethylmalonylcoa',
             }


def sort_genes(genes):
    genes.sort(key=lambda gene: gene.dna_coords[0])
    gene_groups = []
    gene_group = []
    previous_direction = None
    for i, gene in enumerate(genes):
        current_direction = gene.strand
        if current_direction != previous_direction:
            if gene_group:
                if previous_direction == -1:
                    gene_group.reverse()
                gene_groups += gene_group

            gene_group = [gene]
            previous_direction = current_direction
        else:
            gene_group.append(gene)
            if i == len(genes) - 1:
                if current_direction == -1:
                    gene_group.reverse()
                gene_groups += gene_group

    return gene_groups


def parse_antismash_modules(antismash_gbk, screen):
    gene_name_to_genes = {}
    module_nr = -1
    gene_name = None

    for record in SeqIO.parse(antismash_gbk, "genbank"):
        for feature in record.features:
            if feature.type == 'CDS':
                add_gene = False
                module_nr = -1
                qualifiers = feature.qualifiers

                if 'gene' in qualifiers:
                    gene_name = qualifiers['gene'][0]
                elif 'protein' in qualifiers:
                    gene_name = qualifiers['protein'][0]
                elif 'locus_tag' in qualifiers:
                    gene_name = qualifiers['locus_tag'][0]
                else:
                    gene_name = f'gene {len(gene_name_to_genes) + 1}'

                assert gene_name

                if 'gene_functions' in qualifiers:

                    for gene_function in qualifiers['gene_functions']:
                        if gene_function in NRPS_PKS_RULES:
                            add_gene = True

                    if add_gene:
                        gene = Gene(screen, len(gene_name_to_genes))
                        gene.name = gene_name
                        gene.dna_coords = (feature.location.start, feature.location.end)
                        gene.strand = feature.location.strand
                        gene_name_to_genes[gene_name] = gene

            elif feature.type == 'aSModule':

                module_type = feature.qualifiers['type'][0].upper()
                if gene_name in gene_name_to_genes:
                    gene = gene_name_to_genes[gene_name]
                    if module_type == 'NRPS' or module_type == 'PKS':
                        module_nr += 1
                        gene.add_module(module_nr, module_type)
                        for domain in gene.modules[module_nr].domains[:]:
                            gene.modules[module_nr].remove_domain(domain)

                        gene.modules[module_nr].dna_coords = (feature.location.start, feature.location.end)

                        for domain in feature.qualifiers['domains']:
                            if 'PKS_KS' in domain:
                                gene.modules[module_nr].add_domain('KS', domain_label=domain)
                            elif 'PKS_AT' in domain:
                                gene.modules[module_nr].add_domain('AT', domain_label=domain)
                            elif 'PKS_PP' in domain or 'ACP' in domain:
                                if gene.modules[module_nr].type == 'NRPS':
                                    gene.modules[module_nr].add_domain('PCP', domain_label=domain)
                                else:
                                    gene.modules[module_nr].add_domain('ACP', domain_label=domain)
                            elif 'PKS_KR' in domain:
                                gene.modules[module_nr].add_domain('KR', domain_label=domain)
                            elif 'PKS_DH' in domain:
                                gene.modules[module_nr].add_domain('DH', domain_label=domain)
                            elif 'PKS_ER' in domain:
                                gene.modules[module_nr].add_domain('ER', domain_label=domain)
                            elif 'Thioesterase' in domain:
                                gene.modules[module_nr].add_domain('TE', domain_label=domain)
                            elif 'AMP-binding' in domain:
                                gene.modules[module_nr].add_domain('A', domain_label=domain)
                            elif 'Epimerization' in domain:
                                gene.modules[module_nr].add_domain('E', domain_label=domain)
                            elif 'Condensation' in domain or 'Cglyc' in domain:
                                gene.modules[module_nr].add_domain('C', domain_label=domain)
                            elif 'nMT' in domain:
                                gene.modules[module_nr].add_domain('nMT', domain_label=domain)
                            elif 'PCP' in domain:
                                if gene.modules[module_nr].type == 'NRPS':
                                    gene.modules[module_nr].add_domain('PCP', domain_label=domain)
                                else:
                                    gene.modules[module_nr].add_domain('ACP', domain_label=domain)

    return gene_name_to_genes


def parse_antismash_domains(antismash_gbk, gene_name_to_genes):

    gene_name = None
    genes = []

    for record in SeqIO.parse(antismash_gbk, "genbank"):
        for feature in record.features:
            if feature.type == 'CDS':

                qualifiers = feature.qualifiers

                if 'gene' in qualifiers:

                    gene_name = qualifiers['gene'][0]

                elif 'protein' in qualifiers:
                    gene_name = qualifiers['protein'][0]
                elif 'locus_tag' in qualifiers:
                    gene_name = qualifiers['locus_tag'][0]
                else:
                    gene_name = f'gene {len(genes) + 1}'

                assert gene_name
                genes.append(gene_name)

            elif feature.type == 'aSDomain':
                if gene_name in gene_name_to_genes:
                    gene = gene_name_to_genes[gene_name]
                    domains = gene.get_domains()
                    domain_id = feature.qualifiers['domain_id'][0]

                    for domain in domains:
                        if domain.domain_label == domain_id:
                            if 'specificity' in feature.qualifiers:
                                for spec in feature.qualifiers['specificity']:
                                    if 'consensus:' in spec:
                                        specificity = spec.split('consensus:')[1].strip()
                                        if specificity in AS_TO_NRPS:
                                            substrate_name = AS_TO_NRPS[specificity]
                                            substrate_smiles = NRPS_NAME_TO_SMILES[substrate_name]
                                            domain.substrate = Substrate(substrate_name, substrate_smiles, 'NRPS')
                                        elif specificity in AS_TO_PKS:
                                            substrate_name = AS_TO_PKS[specificity]
                                            substrate_smiles = PKS_NAME_TO_SMILES[substrate_name]
                                            domain.substrate = Substrate(substrate_name, substrate_smiles, 'PKS')

                                    elif 'KR stereochemistry:' in spec:
                                        specificity = spec.split('KR stereochemistry:')[1].strip()
                                        if specificity != '(unknown)':
                                            domain.subtype = specificity


def genes_from_antismash(antismash_gbk, screen):
    gene_name_to_genes = parse_antismash_modules(antismash_gbk, screen)
    parse_antismash_domains(antismash_gbk, gene_name_to_genes)

    genes = []

    for gene in gene_name_to_genes.values():
        genes.append(gene)

    genes = sort_genes(genes)
    for i, gene in enumerate(genes):
        gene.gene_number = i
        gene.sort_modules()

    return genes




import os
import pygame

import interactive.flatfiles
import interactive.images.nrps_substrates
import interactive.images.pks_substrates
import interactive.images.pks_starter_substrates

from pikachu.general import png_from_smiles

from interactive.style import SUBSTRATE_BUTTON_SIZE, SUBSTRATE_BUTTON_PADDING, \
    SUBSTRATE_BUTTONS_PER_LINE, WIDTH, HEIGHT, FATTY_ACID_IMAGE_SIZE

SUBSTRATE_DIR = os.path.dirname(interactive.flatfiles.__file__)
NRPS_SUBSTRATE_IMG = os.path.dirname(interactive.images.nrps_substrates.__file__)
PKS_SUBSTRATE_IMG = os.path.dirname(interactive.images.pks_substrates.__file__)
PKS_STARTER_SUBSTRATE_IMG = os.path.dirname(interactive.images.pks_starter_substrates.__file__)
SUBSTRATE_TO_ABBREVIATION_FILE = os.path.join(SUBSTRATE_DIR, 'substrate_to_abbreviation.txt')

PROTEINOGENIC_SUBSTRATES = {"Ala": ['Alanine',
                                    'D-Alanine',
                                    'dehydroalanine',
                                    'beta-Alanine',
                                    '2-methyl-beta-alanine'],
                            "Arg": ["Arginine",
                                    "D-Arginine",
                                    "4,5-dehydroarginine",
                                    "homoarginine"],
                            "Asn": ["Asparagine",
                                    "D-Asparagine"],
                            "Asp": ["AsparticAcid",
                                    "D-AsparticAcid"],
                            "Cys": ["Cysteine",
                                    "D-Cysteine",
                                    "dehydro-cysteine",
                                    "cysteicacid",
                                    "D-cysteicacid",
                                    "alpha-methylcysteine",
                                    "D-alpha-methylcysteine"],
                            "Gln": ["Glutamine",
                                    "D-Glutamine",
                                    "3-Hydroxyglutamine"],
                            "Glu": ["GlutamicAcid",
                                    "D-GlutamicAcid",
                                    "4-methylglutamicacid"],
                            "Gly": ["Glycine"],
                            "His": ["Histidine",
                                    "D-Histidine"],
                            "Ile": ["Isoleucine",
                                    "D-Isoleucine",
                                    "allo-Isoleucine",
                                    "D-allo-Isoleucine",
                                    "homoisoleucine"],
                            "Leu": ["Leucine",
                                    "D-Leucine",
                                    "tert-Leu",
                                    "D-tert-Leucine"],
                            "Lys": ["Lysine",
                                    "D-Lysine",
                                    "beta-lysine"],
                            "Met": ["Methionine",
                                    "D-Methionine"],
                            "Phe": ["Phenylalanine",
                                    "D-Phenylalanine",
                                    "Homophenylalanine",
                                    "N-acetylphenylalanine",
                                    "alpha-amino-phenyl-valericacid",
                                    "D-alpha-amino-phenyl-valericacid"],
                            "Pro": ["Proline",
                                    "D-Proline",
                                    "4-methylproline"],
                            "Ser": ["Serine",
                                    "D-Serine",
                                    "Homoserine",
                                    "D-Homoserine",
                                    "2-Methylserine"],
                            "Thr": ["Threonine",
                                    "D-Threonine",
                                    "allo-Threonine",
                                    "D-allo-Threonine",
                                    "4-butenyl-4-methylthreonine"],
                            "Trp": ["Tryptophan",
                                    "D-Tryptophan",
                                    "2,3-Dehydro-Tryptophan"],
                            "Tyr": ["Tyrosine",
                                    "D-Tyrosine",
                                    "dehydrotyrosine",
                                    "(R)beta-tyrosine",
                                    "Homotyrosine",
                                    "D-Homotyrosine",
                                    "3-methyltyrosine",
                                    "4,5-Dihydroxyhomotyrosine",
                                    "3-hydroxy-4-O-methyl-5-methyltyrosine"],
                            "Val": ["Valine",
                                    "D-Valine",
                                    "4-Hydroxyvaline",
                                    "isovaline",
                                    "D-isovaline",
                                    "alpha-ethylnorvaline",
                                    "D-alpha-ethylnorvaline",
                                    "Norvaline"]}

NONPROTEINOGENIC_SUBSTRATES = {"Aad": ["2-Aminoadipicacid",
                                       "D-2-Aminoadipicacid"],
                               "Abu": ["2-Aminobutyricacid",
                                       "D-2-Aminobutyricacid",
                                       "2-Aminoisobutyricacid",
                                       "Dehydrobutyrine"],
                               "Cit": ["Citrulline",
                                       "D-Citrulline"],
                               "Cor": ["coronamicacid",
                                       "R-coronamicacid",
                                       "norcoronamicacid",
                                       "R-norcoronamicacid"],
                               "Dab": ["2,4-diaminobutyricacid",
                                       "D-2,4-diaminobutyricacid",
                                       "(2S)-2,3-diaminobutyricacid",
                                       "(2S,3S)-2,3-diaminobutyricacid",
                                       "(2R,3R)-2,3-diaminobutyricacid",
                                       "2,3-diamino-3-methyl-propanoicacid",
                                       "D-2,3-Diaminopropionicacid"],
                               "End": ["enduracididine",
                                       "D-enduracididine"],
                               "Kyn": ["kynurenine",
                                       "D-kynurenine"],
                               "Orn": ["Ornithine",
                                       "D-Ornithine",
                                       "4,5-Dihydroxyornithine"],
                               "PheGly": ["phenylglycine",
                                          "D-PhenylGlycine",
                                          "HydroxyPhenylGlycine",
                                          "D-HydroxyPhenylGlycine",
                                          "3,5-dihydroxyphenylglycine",
                                          "D-3,5-dihydroxyphenylglycine"],
                               "Pip": ["Pipecolicacid",
                                       "D-Pipecolicacid",
                                       "L-piperazicacid"],
                               "Other": ["Anticapsin",
                                         "Beta-hydroxycyclohex-2-enylalanine",
                                         "capreomycidine",
                                         "phosphinothricin",
                                         "D-phosphinothricin",
                                         "(2S,3R)-2-amino-3-hydroxy-4-(4-nitrophenyl)butanoate",
                                         "3-(3-Pyridyl)-Alanine",
                                         "3-[(1R,2R)-2-Nitrocyclopropyl]-L-alanine"]}

NON_AMINO_ACIDS = {'Aromatic': ["2-carboxyquinoxaline",
                                "Quinoline-2-carboxylicacid",
                                "2,3-dihydroxybenzoicacid",
                                "4-Hydroxyindole-3-carboxylicacid",
                                "dihydroxyphenylthiazolgroup",
                                "D-dihydroxyphenylthiazolgroup",
                                "D-lysergicacid",
                                "hydrocinnamicacid",
                                "hydroxypicolinicacid",
                                "phenylaceticacid",
                                "salicylicacid"],
                   # 'Cyclic': ["cycloOrnithine",
                   #            "D-cycloOrnithine",
                   #            "homoserinelactone",
                   #            "D-homoserinelactone"],
                   'Other': ["2-hydroxyisovalerate",
                             "D-2-hydroxyisovalerate",
                             # "3-aminopropanimidamide",
                             # "Alaninol",
                             "alpha-ketoisocaproicacid",
                             # "alpha-ketoisovalericacid",
                             "Glycocyamine",
                             "Glycolicacid"]
                             # "valinol"]
                   }

# FATTY_ACIDS = {'Amino group': ['2-Amino-9,10-epoxy-8-oxodecanoidacid',
#                                'D-methyl-2-aminooctanoicacid',
#                                'methyl-2-aminooctanoicacid'],
#
#                'No amino group': [],
#                'Custom': []}

FATTY_ACIDS = {'Amino group': ['methyl-2-aminooctanoicacid',
                               'D-methyl-2-aminooctanoicacid'],
               'Custom': []}


def parse_substrate_to_abbr():
    substrate_to_abbreviation = {}
    with open(SUBSTRATE_TO_ABBREVIATION_FILE, 'r') as abbreviation_file:
        for line in abbreviation_file:
            line = line.strip()
            if line:
                substrate, abbreviation = line.split()
                substrate_to_abbreviation[substrate] = abbreviation

    return substrate_to_abbreviation


SUBSTRATE_TO_ABBREVIATION = parse_substrate_to_abbr()


class SubstrateGroup:
    def __init__(self, name, substrates):
        self.name = name
        self.substrates = substrates
        self.coords = []
        self.create_raster()
        self.set_rectangles()

    def create_raster(self):

        x_coord = int(0.23 * WIDTH)
        y_coord = int(0.52 * HEIGHT)
        column_nr = 0

        for i, substrate in enumerate(self.substrates):
            if i % SUBSTRATE_BUTTONS_PER_LINE == 0:
                x_coord = int(0.23 * WIDTH)
                column_nr += 1
                if i != 0 and column_nr <= 2:
                    y_coord += SUBSTRATE_BUTTON_SIZE + SUBSTRATE_BUTTON_PADDING
                elif i != 0:
                    x_coord = int(0.23 * WIDTH - SUBSTRATE_BUTTON_SIZE - SUBSTRATE_BUTTON_PADDING)

            else:
                x_coord += SUBSTRATE_BUTTON_SIZE + SUBSTRATE_BUTTON_PADDING

            self.coords.append((x_coord, y_coord))

    def set_rectangles(self):

        for i, substrate in enumerate(self.substrates):
            x, y = self.coords[i]
            substrate.x = x
            substrate.y = y
            substrate.rectangle = pygame.Rect(x, y, SUBSTRATE_BUTTON_SIZE, SUBSTRATE_BUTTON_SIZE)


class Substrate:
    def __init__(self, name, smiles, module_type, set_images=True, custom=False, starter=False):
        self.name = name
        self.abbr = self.name
        if self.name in SUBSTRATE_TO_ABBREVIATION:
            self.abbr = SUBSTRATE_TO_ABBREVIATION[self.name]
        self.smiles = smiles
        self.image = None
        self.highlight_image = None
        self.custom = custom
        self.starter = starter

        if set_images:
            if module_type == 'NRPS':
                self.image = os.path.join(NRPS_SUBSTRATE_IMG, f'{name}.png')
                self.highlight_image = os.path.join(NRPS_SUBSTRATE_IMG, f'{name}_highlight.png')

            if module_type == 'PKS':
                self.image = os.path.join(PKS_SUBSTRATE_IMG, f'{name}.png')
                self.highlight_image = os.path.join(PKS_SUBSTRATE_IMG, f'{name}_highlight.png')

            if module_type == 'PKS_starter':
                self.image = os.path.join(PKS_STARTER_SUBSTRATE_IMG, f'{name}.png')
                self.highlight_image = os.path.join(PKS_STARTER_SUBSTRATE_IMG, f'{name}_highlight.png')

        self.rectangle = None
        self.x = 0
        self.y = 0
        self.width = SUBSTRATE_BUTTON_SIZE
        self.height = SUBSTRATE_BUTTON_SIZE


class FattyAcid:
    def __init__(self, c_nr):
        self.c_nr = c_nr
        self.isoform = None
        self.double_bonds = []
        self.tailoring_groups = []
        self.smiles = ''
        self.to_smiles()
        self.image = None
        self.scaled_image = None
        self.draw()
        self.name = None
        self.set_name()

    def set_name(self):
        name = f"C{self.c_nr}"
        if self.isoform:
            name = f"{self.isoform}-{name}"
        if self.double_bonds:
            name = f"{name}:{len(self.double_bonds)}"

        self.name = name

    def set_isoform(self, isoform):
        self.isoform = isoform
        self.draw()

    def set_substrate(self, domain):
        self.to_smiles()
        self.set_name()
        substrate = Substrate(self.name, self.smiles, 'NRPS', set_images=False, custom=True, starter=True)
        domain.substrate = substrate

    def add_methyl_group(self, position):
        self.tailoring_groups.append(('C', position))
        self.draw()

    def add_amino_group(self, position):
        self.tailoring_groups.append(('N', position))
        self.draw()

    def add_hydroxyl_group(self, position):
        self.tailoring_groups.append(('O', position))
        self.draw()

    def add_double_bond(self, position, stereochem):
        self.double_bonds.append((position, stereochem))
        self.draw()

    def count_involvement(self):
        positions = {}
        for pos, cistrans in self.double_bonds:
            if pos not in positions:
                positions[pos] = 0
            positions[pos] += 1

            if pos + 1 not in positions:
                positions[pos + 1] = 0
            positions[pos + 1] += 1

        for group, pos in self.tailoring_groups:
            if pos not in positions:
                positions[pos] = 0

            positions[pos] += 1

        if self.isoform == 'iso':
            if 2 not in positions:
                positions[2] = 0
            positions[2] += 1

        elif self.isoform == 'anteiso':
            if 3 not in positions:
                positions[3] = 0
            positions[3] += 1

        return positions

    def draw(self):
        self.to_smiles()
        path = os.path.join(os.getcwd(), 'temp_fatty_acid_raichu.png')
        png_from_smiles(self.smiles, path)
        self.image = pygame.image.load(path)

        image_width = self.image.get_width()
        image_height = self.image.get_height()

        fatty_acid_image_width, fatty_acid_image_height = FATTY_ACID_IMAGE_SIZE

        if image_width <= fatty_acid_image_width and image_height > fatty_acid_image_height:
            ratio = fatty_acid_image_height / image_height
            image_width *= ratio
            image_height = fatty_acid_image_height
        elif image_width > fatty_acid_image_width and image_height <= fatty_acid_image_height:
            ratio = fatty_acid_image_width / image_width
            image_height *= ratio
            image_width = fatty_acid_image_width
        elif image_width > fatty_acid_image_width and image_height > fatty_acid_image_height:
            ratio_1 = fatty_acid_image_width / image_width
            ratio_2 = fatty_acid_image_height / image_height
            if ratio_1 <= ratio_2:
                ratio = ratio_1
            else:
                ratio = ratio_2

            image_height *= ratio
            image_width *= ratio

        render_window_size = (int(image_width), int(image_height))
        self.scaled_image = pygame.transform.smoothscale(self.image, render_window_size)

    def to_smiles(self):
        smiles = self.c_nr * ['C'] + ['(=O)O']

        if self.isoform == 'iso':
            smiles.pop(0)
            smiles[1] += '(C)'
        elif self.isoform == 'anteiso':
            smiles.pop(0)
            smiles[2] += '(C)'

        self.double_bonds.sort(reverse=True, key=lambda x: x[0])
        for double_bond, stereochem in self.double_bonds:
            smiles[double_bond - 1] += '='

            if stereochem and double_bond != 1:
                carbon_2 = smiles[double_bond]

                if carbon_2.endswith('/') and stereochem == 'trans':
                    smiles[double_bond - 2] += '/'
                elif carbon_2.endswith('/') and stereochem == 'cis':
                    smiles[double_bond - 2] += '\\'
                elif carbon_2.endswith('\\') and stereochem == 'trans':
                    smiles[double_bond - 2] += '\\'
                elif carbon_2.endswith('\\') and stereochem == 'cis':
                    smiles[double_bond - 2] += '/'
                elif stereochem == 'trans':
                    smiles[double_bond] += '/'
                    smiles[double_bond - 2] += '/'
                elif stereochem == 'cis':
                    smiles[double_bond] += '/'
                    smiles[double_bond - 2] += '\\'

        self.tailoring_groups.sort(reverse=True, key=lambda x: x[1])

        for group, pos in self.tailoring_groups:
            carbon = smiles[pos - 1]
            if not carbon.endswith('C'):
                carbon_pos = None
                for i, symbol in enumerate(carbon):
                    if symbol == 'C':
                        carbon_pos = i
                        break
                assert carbon_pos != None
                smiles[pos - 1] = carbon[:carbon_pos + 1] + f'({group})' + carbon[carbon_pos + 1:]

            else:
                smiles[pos - 1] += f'({group})'

        self.smiles = ''.join(smiles)
